#!/usr/bin/env python
'''
Author     : Sunao Sugiyama 
Last edit  : 2024/03/25 21:00:58

Description:
utils.py contains utility functions for fastnc.
'''
import numpy as np
import fcntl
import os

# file locking
def acquire_lock(lock_file):
    lock_fd = open(lock_file, 'w')
    fcntl.flock(lock_fd, fcntl.LOCK_EX)

def release_lock(lock_file):
    lock_fd = open(lock_file, 'w')
    fcntl.flock(lock_fd, fcntl.LOCK_UN)

def npload_lock(filename, suffix='.npz'):
    lock_file = filename.replace(suffix, '.lock')
    acquire_lock(lock_file)
    try:
        data = np.load(filename)
    finally:
        release_lock(lock_file)
    return data

def npsavez_lock(filename, data, suffix='.npz'):
    lock_file = filename.replace(suffix, '.lock')
    acquire_lock(lock_file)
    try:
        np.savez(filename, **data)
    finally:
        release_lock(lock_file)

# Binning utilities
def loglinear(xmin, xmid, xmax, nbin1, nbin2):
    """
    Create a log-linear binning.

    Parameters:
        xmin (float) : The minimum value.
        xmid (float) : The value where the logarithmic and linear ranges meet.
        xmax (float) : The maximum value.
        nbin1 (int)  : The number of bins in the logarithmic range.
        nbin2 (int)  : The number of bins in the linear range.

    Returns:
        numpy.ndarray: An array containing the log-linear binning.
    """
    xbin1 = np.logspace(np.log10(xmin), np.log10(xmid), nbin1+1)[:-1]
    xbin2 = np.linspace(xmid, xmax, nbin2)
    xbin  = np.hstack([xbin1, xbin2])
    return xbin
    
def edge_correction(x, vmin, vmax, atol=1e-8, rtol=1e-5):
    """
    Replace values in x that are out of bounds but close to the boundary edges.
    This is useful for avoiding numerical errors triangle parameter conversion.
    Triangle is parametrized in multiple ways, 

    Parameters:
        x (ndarray)  : Input array.
        vmin (float) : Minimum value.
        vmax (float) : Maximum value.

    Returns:
        out (ndarray): Output array.
    """
    out = x.copy()
    out[np.logical_and(np.isclose(x, vmin, atol, rtol), x<vmin)] = vmin
    out[np.logical_and(np.isclose(x, vmax, atol, rtol), x>vmax)] = vmax
    return out

# Triangle phase factor
def sincos2angbar(psi, delta):
    """
    Calculate triangle phase factor 
    in sine and cosine form.

    Parameters:
        psi (float)   : psi.
        delta (float) : delta.
    """
    cos2b = np.cos(delta) + np.sin(2*psi)
    sin2b = np.cos(2*psi) * np.sin(delta)
    norm  = np.sqrt(cos2b**2 + sin2b**2)
    return sin2b/norm, cos2b/norm

# Config utilities
def merge_config_kwargs(config=None, **kwargs):
    """
    Merge the configuration dictionary and the keyword arguments.

    Parameters:
        config (dict) : Configuration dictionary.
        **kwargs      : Keyword arguments.
    """
    return {**(config or {}), **kwargs}

def get_config_key(config, key, default=None, **kwargs):
    """
    Get the value of a key from the configuration dictionary.

    Parameters:
        config (dict) : Configuration dictionary.
        key (str)     : Key.
        default       : Default value.
        **kwargs      : Keyword arguments.
    """
    config = merge_config_kwargs(config, **kwargs)
    return config.get(key, default)

def update_config(config_base, config=None, **kwargs):
    """
    Update the configuration dictionary.

    Parameters:
        config_base (dict) : Base configuration dictionary.
        config (dict)      : Configuration dictionary.
        **kwargs           : Keyword arguments.
    """
    config = merge_config_kwargs(config, **kwargs)
    config_base.update((key, value) for key, value in config.items() \
            if key in config_base)



"""
Author: Sunao Sugiyama
Last edit: 2024/01/18
"""
import numpy as np
from . import fftlog
from scipy.interpolate import InterpolatedUnivariateSpline as ius
from . import trigutils
from .utils import loglinear

class OneJNaturalConponent:
    """
    See https://arxiv.org/abs/astro-ph/0308328 and https://arxiv.org/abs/2208.11686
    """
    def __init__(self, bispectrum=None):
        if bispectrum is not None:
            self.set_bispectrum(bispectrum)

    def set_bispectrum(self, bispectrum, method='interp'):
        self.bispectrum = bispectrum
        self.method_bispec = method

        self.ellmin = self.bispectrum.ellmin
        self.ellmax = self.bispectrum.ellmax
        self.psimin = self.bispectrum.psimin
        self.psimax = self.bispectrum.psimax
        self.mumin  = self.bispectrum.mumin
        self.mumax  = self.bispectrum.mumax

    def expibarbeta(self, psi, dbeta):
        cos2b = np.cos(dbeta) + np.sin(2*psi)
        sin2b = np.cos(2*psi) * np.sin(dbeta)
        norm  = np.sqrt(cos2b**2 + sin2b**2)
        out = cos2b/norm + 1j*sin2b/norm
        return out**0.5
    
    def expialpha(self, psi, dbeta, tau, dvphi):
        zeta = 0.5*(dbeta-dvphi)
        cosa =  np.cos(psi+tau)*np.sin(zeta)
        sina = -np.cos(psi-tau)*np.cos(zeta)
        norm = np.sqrt(cosa**2 + sina**2)
        eia = cosa/norm + 1j*sina/norm
        return eia

    def Atilde(self, psi, dbeta, tau, dvphi):
        zeta = 0.5*(dbeta-dvphi)
        out = np.cos(psi-tau)**2*np.cos(zeta)**2 + np.cos(psi+tau)**2*np.sin(zeta)**2
        out = np.sqrt(out)
        return out

    def get_f_of_logA(self, psi, dbeta, order=6, extrap=False):
        ell = np.logspace(np.log10(self.ellmin), np.log10(self.ellmax), 256)
        ell1, ell2, ell3 = trigutils.xpsimu_to_x1x2x3(ell, psi, -np.cos(dbeta))
        bs = self.bispectrum.kappa_bispectrum(ell1, ell2, ell3, method=self.method_bispec)
        # extrap for FFTLog
        N_extrap_high = int(0.1*ell.size) if bs[-1] != 0 and extrap else 0
        N_extrap_low  = int(0.1*ell.size) if bs[0] != 0 and extrap else 0
        hankel = fftlog.hankel(ell, ell**4*bs, N_extrap_low=N_extrap_low, N_extrap_high=N_extrap_high, N_pad=200)
        A, f = hankel.hankel(order)
        f_of_logA = ius(np.log(A), A*f, ext=1)
        return f_of_logA

    def integrand_sym(self, t, tau, dvphi, psi, dbeta):
        f6_of_logA = self.get_f_of_logA(psi, dbeta, order=6)
        f2_of_logA = self.get_f_of_logA(psi, dbeta, order=2)

        o0 = 0
        o1 = 0
        o2 = 0
        o3 = 0
        for _psi in [psi, np.pi/2-psi]:
            for _dbeta in [dbeta, -dbeta]:
                Atil = self.Atilde(_psi, _dbeta, tau, dvphi)
                f6 = f6_of_logA(np.log(t*Atil))/(t*Atil)
                f2 = f2_of_logA(np.log(t*Atil))/(t*Atil)

                eb = self.expibarbeta(_psi, _dbeta)
                ea = self.expialpha(_psi, _dbeta, tau, dvphi)
                ez = np.exp(0.5j*(_dbeta-dvphi))
                si = np.sin(2*_psi)

                o0+= si * eb**2 * ea**-6 * f6
                o1+= si * eb**-2* ea**-2 * f2
                o2+= si * eb**2 * ea**-2 * f2 * ez**-4
                o3+= si * eb**2 * ea**-2 * f2 * ez**4

        return o0, o1, o2, o3

    def compute_for_scalar(self, t, tau, phi, projection='x', nbin_psi=100, nbin_dbeta=101):
        """
        t (array)
        tau (float)
        phi (float)

        This takes 90sec for 100x100 grid.
        """
        # psi = loglinear(self.psimin, 1e-3, self.psimax, 60, 60)
        psi = np.linspace(self.psimin, self.psimax, nbin_psi)
        # dbeta = np.pi-np.arccos(1-loglinear(1-self.mumax, 5e-2, 1-self.mumin, 70, 60)[::-1])
        dbeta = np.pi-np.arccos(1-loglinear(1-self.mumax, 5e-2, 1-self.mumin, nbin_dbeta, nbin_dbeta)[::-1])

        dvphi = -phi

        gam0 = []
        gam1 = []
        gam2 = []
        gam3 = []
        for _psi in psi:
            _0 = []
            _1 = []
            _2 = []
            _3 = []
            for _dbeta in dbeta:
                o0, o1, o2, o3 = self.integrand_sym(t, tau, dvphi, _psi, _dbeta)
                gam0.append(o0)
                gam1.append(o1)
                gam2.append(o2)
                gam3.append(o3)
        gam0 = np.reshape(gam0, (len(psi), len(dbeta), -1))
        gam1 = np.reshape(gam1, (len(psi), len(dbeta), -1))
        gam2 = np.reshape(gam2, (len(psi), len(dbeta), -1))
        gam3 = np.reshape(gam3, (len(psi), len(dbeta), -1))
        gam0 = np.trapz(gam0, psi, axis=0)
        gam1 = np.trapz(gam1, psi, axis=0)
        gam2 = np.trapz(gam2, psi, axis=0)
        gam3 = np.trapz(gam3, psi, axis=0)
        gam0 = -np.trapz(gam0, dbeta, axis=0)/2/(2*np.pi)**3
        gam1 = -np.trapz(gam1, dbeta, axis=0)/2/(2*np.pi)**3
        gam2 = -np.trapz(gam2, dbeta, axis=0)/2/(2*np.pi)**3
        gam3 = -np.trapz(gam3, dbeta, axis=0)/2/(2*np.pi)**3

        return gam0, gam1, gam2, gam3

    def compute(self, t, tau, phi, projection='x', nbin_psi=100, nbin_dbeta=101):
        """
        Note:
            running this method can take a really long time.
        """
        shape = t.shape
        t, tau, phi = t.ravel(), tau.ravel(), phi.ravel()

        Gamma0 = np.zeros(t.size, dtype=complex)
        Gamma1 = np.zeros(t.size, dtype=complex)
        Gamma2 = np.zeros(t.size, dtype=complex)
        Gamma3 = np.zeros(t.size, dtype=complex)

        try:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
            size = comm.Get_size()
        except:
            comm = None
            rank = 0
            size = 1

        _, idx = np.unique([tau, phi], axis=1, return_index=True)

        for i in idx[rank::size]:
            w = (tau==tau[i]) & (phi == phi[i])
            g0, g1, g2, g3 = self.compute_for_scalar(t[w], tau[i], phi[i], nbin_psi=nbin_psi, nbin_dbeta=nbin_dbeta)
            Gamma0[w] = g0
            Gamma1[w] = g1
            Gamma2[w] = g2
            Gamma3[w] = g3

        if (comm is not None) and rank==0:
            recvbuf0 = np.empty([size, t.size], dtype=complex)
            recvbuf1 = np.empty([size, t.size], dtype=complex)
            recvbuf2 = np.empty([size, t.size], dtype=complex)
            recvbuf3 = np.empty([size, t.size], dtype=complex)
            comm.Gather(Gamma0, recvbuf0, root=0)
            comm.Gather(Gamma1, recvbuf1, root=0)
            comm.Gather(Gamma2, recvbuf2, root=0)
            comm.Gather(Gamma3, recvbuf3, root=0)
            Gamma0 = np.sum(recvbuf0, axis=0)
            Gamma1 = np.sum(recvbuf1, axis=0)
            Gamma2 = np.sum(recvbuf2, axis=0)
            Gamma3 = np.sum(recvbuf3, axis=0)

        self.Gamma0 = Gamma0.reshape(shape)
        self.Gamma1 = Gamma1.reshape(shape)
        self.Gamma2 = Gamma2.reshape(shape)
        self.Gamma3 = Gamma3.reshape(shape)


#!/usr/bin/env python
'''
Author     : Sunao Sugiyama 
Last edit  : 2024/03/25 17:49:02

Description:
multipole.py contains the Multipole class, 
which computes multipole moments
'''
import numpy as np
from scipy.special import eval_legendre

class MultipoleBase:
    """
    Compute multipole of a function f(x).
    using some basis function indexed by L.
    The maximum multipole must be specified by Lmax.

    Parameters:
        x (array)     : x values
        Lmax (int)    : maximum multipole moment
        method (str)  : method to compute the multipole moments. 
                        options are'gauss-legendre' or 'riemann'.
        verbose (bool): whether to print verbose output
    """
    def __init__(self, x, Lmax, method='gauss-legendre', verbose=True):
        self.x = x
        self.Lmax = Lmax
        self.method = method
        self.verbose = verbose
        self._init_basis_function()

    def _init_basis_function(self):
        """
        Initialize the basis function.
        used to compute the multipoles.

        This allows precomputing the basis function
        if necessary, to save computation time.
        """
        pass

    def _get_basis_function(self, L):
        """
        Return the basis function for multipole indices L

        Parameters:
            L (array): multipole moment
        """
        NotImplementedError

    def __get_linear_interp_coeffs(self, fx, axis=0):
        """
        Get the coefficients for a linear interpolation of y(x).
        The function is interpolated as
        ..math::
            f(x) = a_i x + b_i 
        in each bin.

        Parameters:
            fx (array): fx values evaluated at self.x
            axis (int): axis along which to interpolate
        """
        assert self.x.size == fx.shape[axis], "shape of x and fx must be the same"
        fx = np.moveaxis(fx, axis, -1)
        a = np.diff(fx, axis=-1)/np.diff(self.x)
        b = (fx*np.roll(self.x,-1) - np.roll(fx,-1,axis=-1)*self.x)
        b = np.delete(b, -1, axis=-1)
        b = b/np.diff(self.x)
        a, b = np.moveaxis(a, -1, axis), np.moveaxis(b, -1, axis)
        return a, b

    def decompose(self, fx, L, axis=0):
        """
        Decompose a function f(x) into multipole moments
        using the specified method.

        Parameters:
            fx (array): fx values evaluated at self.x
            L (array): multipole moment
            axis (int): axis along which to decompose
        """
        if self.method == 'gauss-legendre':
            return self.__decompose_gauss_legendre(fx, L, axis=axis)
        elif self.method == 'riemann':
            return self.__decompose_riemann(fx, L, axis=axis)
        else:
            raise ValueError(f"method {self.method} is not supported")

    def __decompose_gauss_legendre(self, fx, L, axis=0):
        """
        Decompose a function f(x) into multipole moments
        using gauss-legendre method with coefficients of
        linear interpolation.

        Parameters:
            fx (array): fx values evaluated at self.x
            L (array) : multipole moment
            axis (int): axis along which to decompose
        """
        a, b = self.__get_linear_interp_coeffs(fx, axis=axis)
        w0, w1 = self._get_basis_function(L)
        out = np.tensordot(w1, a, axes=([1], [axis]))
        out+= np.tensordot(w0, b, axes=([1], [axis]))
        return out

    def __decompose_riemann(self, fx, L, axis=0):
        """
        Decompose a function f(x) into multipole moments
        by approximating the integral with a Riemann sum.
        Note: This is **not** recommended, because this can create 
        artificial high multipole moments, even though the 
        true high multipoles are zeros.

        Parameters:
            fx (array): fx values evaluated at self.x
            L (array) : multipole moment
            axis (int): axis along which to decompose
        """
        w = self._get_basis_function(L)
        out = np.tensordot(w, f, axes=([1], [axis])) * (self.x[1]-self.x[0])
        return out

class MultipoleLegendre(MultipoleBase):
    """
    Multipole decomposition with Legendre polynomials as basis.
    """
    __doc__ += MultipoleBase.__doc__
    # MultipoleLegendre class inherits MultipoleBase class
    def _init_basis_function(self):
        if self.method == 'gauss-legendre':
            Lmax = self.Lmax + 2
            self.__init_legendreP_table(Lmax)
        elif self.method == 'riemann':
            Lmax = self.Lmax
            self.__init_legendreP_table(Lmax)
        else:
            raise ValueError(f"method {self.method} is not supported")

    def _get_basis_function(self, L):
        if self.method == 'gauss-legendre':
            # Note: Here axis is incremented by 1, because
            # L is added as the first axis on top of self.x.shape.
            w0 = 0.5*np.diff(self.__get_legendreP_int0(L), axis=1)
            w1 = 0.5*np.diff(self.__get_legendreP_int1(L), axis=1)
            basis = (w0, w1)
        elif self.method == 'rieamnn':
            basis = (2*L[:,None]+1)/2*self.__get_legendreP(L)
        else:
            raise ValueError(f"method {self.method} is not supported")
        return basis

    # Legendre polynomial P_L(x) related functions
    def __init_legendreP_table(self, Lmax):
        """
        Precompute the Legendre polynomials P_L(x) for all L, 
        to save computation time.

        Parameters:
            Lmax (int): maximum multipole moment
        """
        self.legendreP_table = dict()
        for L in np.arange(Lmax+1):
            p = eval_legendre(L, self.x)
            self.legendreP_table[L] = p
    
    def __get_legendreP(self, L):
        """
        Return legendre polynomial P_L(x)

        Parameters:
            L (array): multipole moment
        """
        if np.isscalar(L):
            L = np.array([L])
        
        out = np.zeros(L.shape+self.x.shape)
        for i, _L in enumerate(L):
            if _L<0:
                continue
            out[i,:] = self.legendreP_table[_L]
        return out

    def __get_legendreP_int0(self, L):
        """
        Return integral of legendre polynomial P_L(x)

        Parameters:
            L (array): multipole moment
        """
        out = self.__get_legendreP(L+1) - self.__get_legendreP(L-1)
        # assign L=0
        out[L==0,:] = self.x
        return out

    def __get_legendreP_int1(self, L):
        """
        Return integral of legendre polynomial P_L(x) with x:
        p_L^1 = (2L+1)\\int dx P_L(x) x

        Parameters:
            L (array): multipole moment
        """
        pL   = self.__get_legendreP(L)
        pLm1 = self.__get_legendreP(L-1)
        pLm2 = self.__get_legendreP(L-2)
        pLp1 = self.__get_legendreP(L+1)
        pLp2 = self.__get_legendreP(L+2)
        out = (pL - pLm2)/(2*L[:,None]-1) \
            + self.x[None,:]*(pLp1 - pLm1) \
            + (pL - pLp2)/(2*L[:,None]+3)
        # assign L=0, 1
        out[L==0,:] = self.x**2/2
        out[L==1,:] = self.x**3
        return out
    
class MultipoleFourier(MultipoleBase):
    """
    Multipole decomposition with Fourier basis as basis.
    """
    __doc__ += MultipoleBase.__doc__
    # MultipoleFourier class inherits MultipoleBase class
    def _init_basis_function(self):
        """
        Fourier basis is computationally low cost,
        so no need to precompute the basis function.
        """
        pass

    def _get_basis_function(self, L):
        if self.method == 'gauss-legendre':
            w0 = np.zeros(L.shape+self.x.shape, dtype=complex)
            w1 = np.zeros(L.shape+self.x.shape, dtype=complex)
            # L == 0
            w0[L==0,:] = self.x[None,:]
            w1[L==0,:] = self.x[None,:]**2/2
            # L != 0
            iL = 1j*L[L!=0,None]
            w0[L!=0,:] = np.exp(iL*self.x[None,:])/iL
            w1[L!=0,:] = (self.x[None,:]-1.0/iL)/iL * np.exp(iL*self.x[None,:])
            # diff
            w0 = np.diff(w0, axis=1)
            w1 = np.diff(w1, axis=1)
            basis = (w0, w1)
        elif self.method == 'rieamnn':
            basis = np.exp(1j*L[:,None]*self.x[None,:])
        else:
            raise ValueError(f"method {self.method} is not supported")
        return basis

class MultipoleSine(MultipoleBase):
    """
    Multipole decomposition with sine basis as basis.
    """
    __doc__ += MultipoleBase.__doc__
    # MultipoleSine class inherits MultipoleBase class
    def _init_basis_function(self):
        """
        Sin basis is computationally low cost,
        so no need to precompute the basis function.
        """
        pass

    def _get_basis_function(self, L):
        if self.method == 'gauss-legendre':
            w0 = np.zeros(L.shape+self.x.shape)
            w1 = np.zeros(L.shape+self.x.shape)
            # L == 0
            w0[L==0,:] = 0
            w1[L==0,:] = 0
            # L != 0
            _L, _x = L[L!=0,None], self.x[None,:]
            w0[L!=0,:] = -np.cos(_L*_x)/_L
            w1[L!=0,:] = -_x*np.cos(_L*_x)/_L + np.sin(_L*_x)/_L**2
            # diff
            w0 = np.diff(w0, axis=1)
            w1 = np.diff(w1, axis=1)
            basis = (w0, w1)
        elif self.method == 'rieamnn':
            basis = np.sin(L[:,None]*self.x[None,:])
        else:
            raise ValueError(f"method {self.method} is not supported")
        return basis

class MultipoleCosine(MultipoleBase):
    """
    Multipole decomposition with cosine basis as basis.
    """
    __doc__ += MultipoleBase.__doc__
    # MultipoleCosine class inherits MultipoleBase class
    def _init_basis_function(self):
        """
        Cos basis is computationally low cost,
        so no need to precompute the basis function.
        """
        pass

    def _get_basis_function(self, L):
        if self.method == 'gauss-legendre':
            w0 = np.zeros(L.shape+self.x.shape)
            w1 = np.zeros(L.shape+self.x.shape)
            # L == 0
            w0[L==0,:] = self.x[None,:]
            w1[L==0,:] = self.x[None,:]**2/2
            # L != 0
            _L, _x = L[L!=0,None], self.x[None,:]
            w0[L!=0,:] = np.sin(_L*_x)/_L
            w1[L!=0,:] = _x*np.sin(_L*_x)/_L - np.cos(_L*_x)/_L**2
            # diff
            w0 = np.diff(w0, axis=1)
            w1 = np.diff(w1, axis=1)
            basis = (w0, w1)
        elif self.method == 'rieamnn':
            basis = np.cos(L[:,None]*self.x[None,:])
        else:
            raise ValueError(f"method {self.method} is not supported")
        return basis



__author__ = 'Sunao Sugiyama, Rafael Heringer Gomes'
__version__ = '1.1.17'
__url__ = 'https://github.com/git-sunao/fastnc'
# core modules:
from . import fastnc
from . import multipole
from . import bispectrum
from . import coupling
# supplemental modules:
from . import halofit
from . import trigutils
from . import utils
from . import onejnc
from . import devtool



#!/usr/bin/env python
'''
Author     : Sunao Sugiyama 
Last edit  : 2024/04/08 18:12:59

Description:
halofit.py contains the Halofit class. 
See the references below:
https://arxiv.org/abs/1208.2701
https://arxiv.org/abs/1911.07886
'''
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as ius
from scipy.integrate import simpson
from scipy.optimize import bisect
from scipy.special import gamma, gammaincc

class Halofit:
    """
    Halofit class including power and bi spectrum.
    Fitting formula of the nonlinear matter power spectrum 
    based on the halo model. 

    Parameters:
        k     (np.ndarray): array of comoving Fourier modes
        pklin (np.ndarray): array of linear power spectrum at the wave numbers
        z     (np.ndarray): array of redshifts
        lgr   (np.ndarray): linear growth rate at the redshifts

    References:
        See https://arxiv.org/abs/1208.2701 for halofit model.
        See https://arxiv.org/abs/1911.07886 for bihalofit model.
    """
    def __init__(self, k=None, pklin=None, z=None, lgr=None, cosmo=None):
        self.set_lgr(z, lgr)
        self.set_pklin(k, pklin)
        self.set_cosmology(cosmo)

    def set_lgr(self, z, lgr):
        """
        Set the linear growth rate at the redshifts.

        Parameters:
            z     (np.ndarray): array of redshifts
            lgr   (np.ndarray): linear growth rate at the redshifts
        """
        self.z     = z
        self.lgr   = lgr
        self.has_changed = True
    
    def set_pklin(self, k, pklin):
        """
        Set the linear power spectrum at the wave numbers.

        Parameters:
            k     (np.ndarray): array of comoving Fourier modes
            pklin (np.ndarray): array of linear power spectrum at the wave numbers
        """
        self.k     = k
        self.pklin = pklin
        self._normalize_pklin()
        self.has_changed = True

    def set_cosmology(self, cosmo):
        """
        Set the cosmological parameters.

        Parameters:
            cosmo (dict): dictionary of cosmological parameters

        cosmo needs to have the following keys:
            Om0   : matter density parameter at z=0
            Ode0  : dark energy density parameter at z=0
            ns    : spectral index of linear matter power spectrum
            w0    : w0 parameter of dark energy EoS
            sigma8: sigma8 at z=0
            (below is irrelevant to bihalofit)
            wa    : wa parameter of dark eneryg EoS
            fnu0  : fraction of neutrino energy density relative to matter at z=0
        """
        self.cosmo = cosmo
        self.has_changed = True

    def update(self):
        """
        Update the internal variables.
        """
        if self.has_changed:
            self._normalize_pklin()
            self._init_spline()
            self.has_changed = False
    
    def _normalize_pklin(self):
        """
        Normalize the linear power spectrum with sigma8.
        """
        if (self.k is not None) and (self.pklin is not None) and (self.cosmo is not None):
            # compute sigma8 with current pklin
            k     = np.logspace(-3, 2, 1000)
            Delta = self.get_interpolated_pklin(k)*k**3 / 2/np.pi**2
            sigma8temp = self._sigmam(k, Delta, 8.0, window_tophat)
            self.pklin *= (self.cosmo['sigma8']/sigma8temp)**2

    def get_interpolated_pklin(self, k, z=None, ext=True):
        """
        Interpolate the power spectrum for a given redshift.

        Parameters:
            k   (np.ndarray): array of comoving Fourier modes
            z   (float)     : redshift
            ext (bool)      : if True, extrapolate the power spectrum for higher k
        """
        # array to store log(pk)
        pk = np.zeros_like(k)

        # do interpolation
        where = (self.k.min() <= k) & (k <= self.k.max())
        logpk = ius(np.log(self.k), np.log(self.pklin))(np.log(k[where]))
        pk[where] = np.exp(logpk)

        if ext:
            # extrapolate by power law for lower k
            where = k < self.k.min()
            n = np.log(self.pklin[1]/self.pklin[0]) / np.log(self.k[1]/self.k[0])
            a = self.pklin[0]
            pk[where] = a*(k[where]/self.k[0])**n

            # extrapolate by power law for higher k
            where = k > self.k.max()
            n = np.log(self.pklin[-1]/self.pklin[-2]) / np.log(self.k[-1]/self.k[-2])
            a = self.pklin[-1]
            pk[where] = a*(k[where]/self.k[-1])**n
        
        # multiply the linear growth rate
        # to the linear power at z = 0.0
        if z is not None:
            pk *= ius(self.z, self.lgr, ext=2)(z)**2

        return pk

    def _sigmam(self, k, Delta, r, window, extrap=False):
        """
        Calculate sigmaM.

        Parameters:
            k (np.ndarray)    : array of comoving Fourier modes
            Delta (np.ndarray): array of Delta^2(k)
            r (float)         : smoothing scale
            window (function) : window function
            extrap (bool)     : if True, extrapolate the power spectrum for higher k

        Returns:
            sigmam (float): sigmaM

        Note:
            The integral is calculated by the Simpson's rule.
            If extrap is True, the power spectrum is extrapolated by a power law: A k^n.

            And the integral is calculated as follows:

            ..math::
                sigma_m(R)  = \\int_-inf^inf dlnk Delta(k) W^2(kR)
                            = \\int_-inf^kmax dlnk Delta(k) W^2(kR) + \\int_kmax^inf dlnk Delta(k) W^2(kR)
                            ~ sum_i Delta_i W^2(k_i R) dlnk + \\int_kmax^inf dlnk Delta(k) W^2(kR)

            Assuming the Delta at high k can be approximated by a power law, ~ A k^n, and Gaussian window,

            ..math::
                sigma_m(R)  = I2 + \\int_kmax^inf dlnk Ak^n e^(-k^2R^2)
                            = I2 + A R^(-n) Gamma(n/2) Gamma^reg(n/2, tmin)

            where tmin = kmax^2R^2, and Gamma^reg is the regularized gamma function.
        """
        I2    = simpson(Delta * window(k*r)**2, np.log(k))

        if extrap:
            # spectral index and amplitude of Delta at high k
            n = np.diff(np.log(Delta))[-1]/np.diff(np.log(k))[-1]
            A = Delta[-1] * k[-1]**(-n)
            tmin = k[-1] * r
            I2 += A * r**-n * 0.5*gamma(n/2) * gammaincc(n/2, tmin**2)
            
        return I2**0.5
        
    def get_r_sigma(self, z, rtol=1e-5, maxiter=10):
        """
        Returns nonlinear scale, r_sigma.

        Parameters:
            z       (float): redshift
            rtol    (float): relative tolerance
            maxiter (int)  : maximum number of iterations
        """
        k     = np.logspace(-3, 2, 1000)
        Delta = self.get_interpolated_pklin(k, z)*k**3 / 2/np.pi**2
        def eq(R): return self._sigmam(k, Delta, R, window_gaussian, extrap=True) - 1.0
        # initial guess
        rmin, rmax = 1.0/k.max(), 1.0/k.min()
        niter = 0
        while eq(rmin)*eq(rmax) > 0 and niter <= maxiter:
            rmin /= 2.0
            niter += 1
        if niter > maxiter:
            r_sigma = rmin
        else:
            r_sigma = abs(bisect(eq, rmin, rmax, rtol=rtol))
        return r_sigma
    
    # ingredients
    def get_neff(self, z, r_sigma=None):
        """
        Returns effective spectral index.

        Parameters:
            z       (float): redshift
            r_sigma (float): smoothing scale
        """
        if r_sigma is None:
            r_sigma = self.get_r_sigma(z)
        # Get effecitive spectrum index: neff
        k       = np.logspace(-4, 2, 2000)
        Delta   = self.get_interpolated_pklin(k, z)*k**3 / 2/np.pi**2
        sigmaG1 = self._sigmam(k, Delta, r_sigma, window_gaussian_1deriv)
        neff    = -3.0 + 2.0*sigmaG1**2
        return neff
    
    def get_C(self, z, r_sigma=None):
        """
        Returns spectral curvature.

        Parameters:
            z       (float): redshift
            r_sigma (float): smoothing scale
        """
        if r_sigma is None:
            r_sigma = self.get_r_sigma(z)
        # Get effecitive spectrum index: neff
        k       = np.logspace(-4, 2, 2000)
        Delta   = self.get_interpolated_pklin(k, z)*k**3 / 2/np.pi**2
        sigmaG1 = self._sigmam(k, Delta, r_sigma, window_gaussian_1deriv)
        sigmaG2 = self._sigmam(k, Delta, r_sigma, window_gaussian_2deriv)
        C       = 4*sigmaG1**4 + 4*sigmaG1**2 - 4*sigmaG2**2
        return C
    
    def get_sigma8z(self, z):
        """
        Returns sigma8 at a given redshift z.

        Parameters:
            z (float): redshift
        """
        k       = np.logspace(-4, 2, 1000)
        Delta   = self.get_interpolated_pklin(k, z)*k**3 / 2/np.pi**2
        sigma8z = self._sigmam(k, Delta, 8.0, window_tophat)
        return sigma8z
    
    def get_Omz(self, z):
        """
        Returns Omega_m(z).

        Parameters:
            z (float): redshift
        """
        Om0, Ode0, w0, wa = self.cosmo['Om0'], self.cosmo['Ode0'], self.cosmo['w0'], self.cosmo['wa']
        a = 1.0/(1+z)
        Qa2 = a**(-1.0-3.0*(w0+wa))*np.exp(-3.0*(1-a)*wa)
        Omt =1.0+(Om0+Ode0-1.0)/(1-Om0-Ode0+Ode0*Qa2+Om0/a)
        Omz =Omt*Om0/(Om0+Ode0*a*Qa2)
        return Omz
    
    def get_Odez(self, z):
        """
        Returns Omega_de(z).

        Parameters:
            z (float): redshift
        """
        Om0, Ode0, w0, wa = self.cosmo['Om0'], self.cosmo['Ode0'], self.cosmo['w0'], self.cosmo['wa']
        a = 1.0/(1+z)
        Qa2 = a**(-1.0-3.0*(w0+wa))*np.exp(-3.0*(1-a)*wa)
        Ot =1.0+(Om0+Ode0-1.0)/(1-Om0-Ode0+Ode0*Qa2+Om0/a)
        Ode=Ot*Ode0*Qa2/(Ode0*Qa2+Om0/a)
        return Ode
    
    def _init_spline(self, zmid=0.5, dz_low=0.15, dz_high=0.3):
        """
        Initialize the spline for halofit.

        Parameters:
            zmid    (float): the middle redshift for the spline
            dz_low  (float): the lower redshift interval for the spline
            dz_high (float): the higher redshift interval for the spline
        """
        # For acculate calculation, we divide z array into two sections
        # Defining zmid, the first section is [0, zmid) and the second is [zmid, max),
        # where we use finer bin for the first than the second.
        # The default zmid, dz_low, dz_high is optimized to have 0.05% accuracy 
        # only with ~100msec CPU time.
        if zmid <= self.z.min():
            z = np.arange(self.z.min(), self.z.max(), dz_high)
        elif self.z.max() <= zmid:
            z = np.arange(self.z.min(), self.z.max(), dz_low)
        else:
            z = np.hstack([np.arange(self.z.min(), zmid, dz_low), 
                           np.arange(zmid, self.z.max(), dz_high)])
        if not self.z.max() in z:
            z = np.append(z, self.z.max())
        self.lazy_arrays = {'z':z}
        
        names = ['r_sigma', 'neff', 'C', 'sigma8z', 'Omz', 'Odez']
        for name in names:
            self.lazy_arrays[name] = np.empty(z.shape)
        for i, _z in enumerate(z):
            self.lazy_arrays['r_sigma'][i] = self.get_r_sigma(_z)
            self.lazy_arrays['neff'][i]    = self.get_neff(_z, self.lazy_arrays['r_sigma'][i])
            self.lazy_arrays['C'][i]       = self.get_C(_z, self.lazy_arrays['r_sigma'][i])
            self.lazy_arrays['sigma8z'][i] = self.get_sigma8z(_z)
        self.lazy_arrays['Omz']     = self.get_Omz(z)
        self.lazy_arrays['Odez']    = self.get_Odez(z)
    
    def get_halofit_coeffs(self, z):
        """
        Returns the coefficients of halofit.

        Parameters:
            z (np.ndarray): array of redshifts
        """
        r_sigma = ius(self.lazy_arrays['z'], self.lazy_arrays['r_sigma'])(z)
        neff    = ius(self.lazy_arrays['z'], self.lazy_arrays['neff'])(z)
        C       = ius(self.lazy_arrays['z'], self.lazy_arrays['C'])(z)
        Odez    = ius(self.lazy_arrays['z'], self.lazy_arrays['Odez'])(z)
        Omz     = ius(self.lazy_arrays['z'], self.lazy_arrays['Omz'])(z)
        w       = self.cosmo['w0'] + self.cosmo['wa']*z/(z+1)
        fnu0    = self.cosmo['fnu0']
        
        an = 10.**(1.5222 + 2.8553*neff + 2.3706*neff**2 + 0.9903*neff**3 + 0.2250*neff**4 \
                    - 0.6038*C + 0.1749*Odez*(1+w))
        bn = 10.**(-0.5642 + 0.5864*neff + 0.5716*neff**2 - 1.5474*C + 0.2279*Odez*(1+w))
        cn = 10.**(0.3698 + 2.0404*neff + 0.8161*neff**2 + 0.5869*C)

        gan = 0.1971 - 0.0843*neff + 0.8460*C
        aln = np.abs( 6.0835 + 1.3373*neff - 0.1959*neff**2 - 5.5274*C )
        ben = 2.0379 - 0.7354*neff + 0.3157*neff**2 + 1.2490*neff**3 + 0.3980*neff**4 \
                    - 0.1682*C + fnu0*(1.081 + 0.395*neff**2)
        mun = np.zeros(z.shape)
        nun = 10**(5.2105 + 3.6902*neff)
        
        f1b, f2b, f3b = Omz**-0.0307, Omz**-0.0585, Omz**0.0743
        f1a, f2a, f3a = Omz**-0.0732, Omz**-0.1423, Omz**0.0725
        frac = Odez/(1-Omz)
        f1, f2, f3 = frac*f1b+(1-frac)*f1a, frac*f2b+(1-frac)*f2a, frac*f3b+(1-frac)*f3a
        
        coeffs = dict()
        coeffs['r_sigma'] = r_sigma
        coeffs['an']      = an
        coeffs['bn']      = bn
        coeffs['cn']      = cn
        coeffs['gan']     = gan
        coeffs['aln']     = aln
        coeffs['ben']     = ben
        coeffs['mun']     = mun
        coeffs['nun']     = nun
        coeffs['f1']      = f1
        coeffs['f2']      = f2
        coeffs['f3']      = f3
        
        return coeffs
        
    def get_pkhalofit(self, k, z):
        """
        Returns the halofit prediction of nonlinear matter power spectrum.

        Parameters:
            k (np.ndarray): array of comoving Fourier modes
            z (np.ndarray): array of redshifts
        """
        # update the internal variables
        self.update()

        z = np.asarray(z)
        c = self.get_halofit_coeffs(z)
        fnu0 = self.cosmo['fnu0']
        Om0  = self.cosmo['Om0']
        
        pkhalo = []
        for i, _z in enumerate(z):
            pklin = self.get_interpolated_pklin(k, _z)
            DeltaL= pklin * k**3/2/np.pi**2
            
            y     = k * c['r_sigma'][i]
            f     = y/4. + y**2/8.
            
            DeltaLaa = DeltaL*(1.+fnu0*47.48*k**2/(1.+1.5*k**2))
            DeltaQ   = DeltaL*((1.+DeltaLaa)**c['ben'][i])/(1.+c['aln'][i]*DeltaLaa) * np.exp(-f)
            DeltaH   = c['an'][i]*y**(3.*c['f1'][i]) \
                        / (1.+c['bn'][i]*y**c['f2'][i] + (c['cn'][i]*y*c['f3'][i])**(3.-c['gan'][i]))
            DeltaH   = DeltaH / (1. + c['mun'][i]/y + c['nun'][i]/y**2) \
                        * (1+fnu0*(0.977-18.015*(Om0-0.3)))
            pkhalo.append( (DeltaQ + DeltaH) * (2.*np.pi**2) / k**3 )
        pkhalo = np.array(pkhalo)
    
        return pkhalo
    
    def get_bihalofit_coeffs(self, z):
        """
        Returns the coefficients of bihalofit, 
        or parts of coefficients which are independent of triangle.

        Parameters:
            z (np.ndarray): array of redshifts
        """
        r_sigma = ius(self.lazy_arrays['z'], self.lazy_arrays['r_sigma'])(z)
        neff    = ius(self.lazy_arrays['z'], self.lazy_arrays['neff'])(z)
        Omz     = ius(self.lazy_arrays['z'], self.lazy_arrays['Omz'])(z)
        sigma8z = ius(self.lazy_arrays['z'], self.lazy_arrays['sigma8z'])(z)
        log10sigma8z = np.log10(sigma8z)
        
        # initialize the array
        coeffs = np.zeros(z.shape,dtype=[('r_sigma', 'f8'), 
                                         ('log10an1', 'f8'),
                                         ('log10an2', 'f8'),
                                         ('bn', 'f8'),
                                         ('cn', 'f8'),
                                         ('log10aln1', 'f8'),
                                         ('log10aln2', 'f8'),
                                         ('log10ben1', 'f8'),
                                         ('log10ben2', 'f8'),
                                         ('gan', 'f8'),
                                         ('fn', 'f8'),
                                         ('gn', 'f8'),
                                         ('hn', 'f8'),
                                         ('mn', 'f8'),
                                         ('nn', 'f8'),
                                         ('mun', 'f8'),
                                         ('nun', 'f8'),
                                         ('pn', 'f8'),
                                         ('en', 'f8'),
                                         ('dn', 'f8')])
        coeffs['r_sigma']   = r_sigma
        
        # 1-halo term coefficients
        ## a part of an, add -0.310*r1^gan later
        coeffs['log10an1'] = -2.167-2.944*log10sigma8z-1.106*log10sigma8z**2-2.865*log10sigma8z**3
        coeffs['log10an2'] = -0.310*np.ones(z.shape)
        coeffs['bn'] = 10.**(-3.428-2.681*log10sigma8z+1.624*log10sigma8z**2-0.095*log10sigma8z**3)
        coeffs['cn'] = 10.**(0.159-1.107*neff)
        ## parts of alphan to combined with r2 later
        coeffs['log10aln1'] = -4.348-3.006*neff-0.5745*neff**2
        coeffs['log10aln2'] = 10**(-0.9+0.2*neff) 
        ## parts of betan to be combined with r2 later
        coeffs['log10ben1'] = -1.731-2.845*neff-1.4995*neff**2-0.2811*neff**3
        coeffs['log10ben2'] = 0.007*np.ones(z.shape)
        # gan used in an
        coeffs['gan']  = 10**(0.182+0.57*neff)
        
        # 3-halo term coefficients
        coeffs['fn'] = 10**(-10.533-16.838*neff-9.3048*neff**2-1.8263*neff**3)
        coeffs['gn'] = 10**(2.787+2.405*neff+0.4577*neff**2)
        coeffs['hn'] = 10**(-1.118-0.394*neff)
        coeffs['mn'] = 10**(-2.605-2.434*log10sigma8z+5.71*log10sigma8z**2)
        coeffs['nn'] = 10**(-4.468-3.08*log10sigma8z+1.035*log10sigma8z**2)
        coeffs['mun']= 10**(15.312+22.977*neff+10.9579*neff**2+1.6586*neff**3)
        coeffs['nun']= 10**(1.347+1.246*neff+0.4525*neff**2)
        coeffs['pn'] = 10**(0.071-0.433*neff)
        coeffs['en'] = 10**(-0.632+0.646*neff)
        coeffs['dn'] = 10**(-0.483+0.892*log10sigma8z-0.086*Omz)
        
        return coeffs
        
    def F2_tree(self, k1, k2, k3):
        """
        Returns the tree level bispectrum.

        Parameters:
            k1 (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            k2 (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            k3 (np.ndarray): array of comoving Fourier modes in h/Mpc unit
        """
        costheta12=0.5*(k3*k3-k1*k1-k2*k2)/(k1*k2)
        return (5./7.)+0.5*costheta12*(k1/k2+k2/k1)+(2./7.)*costheta12*costheta12
        
    def get_bihalofit(self, k1, k2, k3, z, verbose=False, all_physical=False, which=['Bh3', 'Bh1']):
        """
        Returns the bihalofit prediction of matter bispectrum.

        Parameters:
            k1           (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            k2           (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            k3           (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            z            (np.ndarray): array of redshifts
            verbose      (bool)      : if True, print the progress
            all_physical (bool)      : if True, assumes all the triangle satisfy the 
                                       triangle condition
            which        (str or list): which part of bispectrum to calculate
        """
        if isinstance(which, str):
            which = [which]

        for term in which:
            if term not in ['BT', 'Bh1', 'Bh3']:
                raise ValueError('which arg should be one of BT, Bh1, Bh3 or a list of them.')

        # update the internal variables
        self.update()
        
        # N
        N = z.shape
        
        # ratio of modes
        kmin, kmid, kmax = np.sort([k1,k2,k3], axis=0)
        r1, r2 = kmin/kmax, (kmid+kmin-kmax)/kmax
        
        # coefficients
        c  = self.get_bihalofit_coeffs(z)
        ns = self.cosmo['ns']
        
        # select physical modes
        sel = kmax <= kmin+kmid
        if all_physical:
            sel[:] = True
        z = z[sel]
        k1, k2, k3 = k1[sel], k2[sel], k3[sel]
        r1, r2 = r1[sel], r2[sel]
        c = c[sel]
        
        # dimensionless Fourier modes
        q1, q2, q3 = k1*c['r_sigma'], k2*c['r_sigma'], k3*c['r_sigma']
        
        # linear power spectrum
        if 'BT' in which or 'Bh3' in which:
            PL1 = self.get_interpolated_pklin(k1, z)
            PL2 = self.get_interpolated_pklin(k2, z)
            PL3 = self.get_interpolated_pklin(k3, z)

        # total bispectrum
        Btot  = np.zeros(N)

        # tree level
        if 'BT' in which:
            BT  = np.zeros(N)
            BT[sel] = 2*self.F2_tree(k1,k2,k3)*PL1*PL2 \
                            + 2*self.F2_tree(k2,k3,k1)*PL2*PL3 \
                            + 2*self.F2_tree(k3,k1,k2)*PL3*PL1
            BT[np.logical_not(sel)] = np.nan
            Btot += BT
        
        # 1-halo term, Eq. (B4)
        if 'Bh1' in which:
            Bh1 = np.ones(N)
            for q in [q1, q2, q3]:
                # Combine parts of coefficients
                an = 10**(c['log10an1' ]+c['log10an2' ]*r1**c['gan'])
                aln= 10**(c['log10aln1']+c['log10aln2']*r2**2)
                aln[aln > 1.-(2./3.)*ns] = 1.-(2./3.)*ns
                bn = c['bn']
                ben= 10**(c['log10ben1']+c['log10ben2']*r2)
                cn = c['cn']
                # Multiply the result to form Bh1
                Bh1[sel] *= 1/(an*q**aln + bn*q**ben) / (1+1/cn/q)
            Bh1[np.logical_not(sel)] = np.nan
            Btot += Bh1
        
        # 3-halo term
        if 'Bh3' in which:
            PE = []
            for PL, q in [(PL1,q1), (PL2,q2), (PL3,q3)]:
                PE.append( (1+c['fn']*q**2)/(1+c['gn']*q+c['hn']*q**2)*PL \
                        + 1/(c['mn']*q**c['mun']+c['nn']*q**c['nun']) \
                        /(1+(c['pn']*q)**-3) )
            PE1, PE2, PE3 = PE
            Bh3 = np.zeros(N)
            Bh3[sel] = 2*(self.F2_tree(k1,k2,k3)+c['dn']*q3)*PE1*PE2 \
                            + 2*(self.F2_tree(k2,k3,k1)+c['dn']*q1)*PE2*PE3 \
                            + 2*(self.F2_tree(k3,k1,k2)+c['dn']*q2)*PE3*PE1
            Bh3[sel]*= 1./(1.+c['en']*q1)/(1.+c['en']*q2)/(1.+c['en']*q3)
            Bh3[np.logical_not(sel)] = np.nan
            Btot += Bh3
        
        return Btot
    
    def get_Rb_bihalofit(self, k1, k2, k3, z):
        """
        Returns the baryon ratio on bispectrum.

        Parameters:
            k1           (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            k2           (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            k3           (np.ndarray): array of comoving Fourier modes in h/Mpc unit
            z            (np.ndarray): array of redshifts
        """
        # ratio of modes
        kmin, kmid, kmax = np.sort([k1,k2,k3], axis=0)
        
        # Baryon effective redshift
        sel = z<5
        
        # coefficients
        a = 1/(1+z[sel])
        A0  = np.zeros(a.shape)
        A0[a>0.5] = 0.068*(a[a>0.5]-0.5)**0.47
        mu0 = 0.018*a + 0.837*a**2
        si0 = 0.881*mu0
        al0 = 2.346
        A1  = np.zeros(a.shape)
        A1[a>0.2] = 1.052*(a[a>0.2]-0.2)**1.41
        mu1 = np.abs(0.172+3.048*a-0.675*a**2)
        si1 = (0.494-0.039*a)*mu1
        kst = 29.90 - 38.73*a+24.30*a**2
        al2 = 2.25
        be2 = 0.563/((a/0.06)**0.02+1) / al2

        # Assign baryon ratios
        Rb = np.ones(z.shape)
        for k in [k1[sel], k2[sel], k3[sel]]:
            x = np.log10(k)
            Rb[sel]*= A0*np.exp(-np.abs((x-mu0)/si0)**al0) \
                        - A1*np.exp(-((x-mu1)/si1)**2) \
                        + ((k/kst)**al2+1)**be2
                
        return Rb
    
def window_tophat(x):
    """
    Top-hat window function.
    """
    return 3.0/x**3 * (np.sin(x) - x*np.cos(x))

def window_gaussian(x):
    """
    Gaussian window function.
    """
    return np.exp(-0.5*x**2)

def window_gaussian_1deriv(x):
    """
    First derivative of Gaussian window function.
    """
    return x*np.exp(-0.5*x**2)

def window_gaussian_2deriv(x):
    """
    Second derivative of Gaussian window function.
    """
    return x**2*np.exp(-0.5*x**2)



#!/usr/bin/env python
'''
Author     : Sunao Sugiyama 
Last edit  : 2024/06/08 00:21:40

Description:
bispectrum.py contains classes for computing bispectrum 
and various methods of bispectrum: interpolation, 
multipole decomposition, etc.
'''
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as ius
from scipy.interpolate import RegularGridInterpolator as rgi
from astropy.cosmology import wCDM, Planck18
from scipy.special import sici
from scipy.special import eval_legendre
# fastnc modules
from . import trigutils
from .halofit import Halofit
from .multipole import MultipoleLegendre, MultipoleFourier
from .utils import loglinear, edge_correction, update_config, get_config_key


wPlanck18 = wCDM(H0=Planck18.H0, Om0=Planck18.Om0, Ode0=Planck18.Ode0, w0=-1.0, meta=Planck18.meta, name='wPlanck18')

class BispectrumBase:
    r"""
    Base class for bispectrum computation.

    Parameters:
        config (dict)       : A configuration dict that can be used to pass in the below kwargs if
                              desired.  This dict is allowed to have addition entries in addition
                              to those listed below, which are ignored here. (default: None)
        verbose (bool)      : Whether to print the progress. (default: True)

    Keyword arguments:
        ell1min (float)     : minimum of ell1 and ell2
        ell1max (float)     : maximum of ell1 and ell2
        epmu (float)        : small number to avoid the squeezed limit
        zmin (float)        : minimum of redshift
        zmid (float)        : middle of redshift to switch from log to linear binning
        nzbin_log (int)     : number of bins for redshift in log scale
        nzbin_lin (int)     : number of bins for redshift in linear scale
        nrbin (int)         : number of bins for r
        nubin (int)         : number of bins for u
        nvbin (int)         : number of bins for v
        method (str)        : method for interpolation
        use_interp (bool)   : whether to use interpolation
        nellbin (int)       : number of bins for ell
        npsibin (int)       : number of bins for psi
        nmubin (int)        : number of bins for mu
        Lmax (int)          : maximum multipole (default: None)
        Lmax_diag (nint)    : maximum multipole for diagonal elements (default: None)
        multipole_type (str): type of multipole decomposition
        method (str)        : method for multipole evaluation
    
    Usage:
    >>> b = BispectrumBase()
    >>> b.set_cosmology(cosmo)
    >>> b.set_source_distribution(zs, pzs)
    >>> b.set_ell1mu_range(ell1min, ell1max, epmu)
    >>> b.interpolate(scombs=None, **args)
    >>> b.decompose(scombs=None, method_bispec='interp', **args)
    >>> b.kappa_bispectrum_multipole(L, ell, psi, scomb=scomb)
    """
    # default configs
    config_scale     = dict(ell1min=None, ell1max=None, epmu=1e-7)
    config_losint    = dict(zmin=1e-4, zmid=1e-1, nzbin_log=15, nzbin_lin=40)
    config_interp    = dict(nrbin=35, nubin=35, nvbin=25, method='linear', use_interp=True)
    config_multipole = dict(nellbin=100, npsibin=80, nmubin=50, nmubin_log=30, Lmax=None, Lmax_diag=None, \
        multipole_type='legendre', method='gauss-legendre')
    config_IA        = dict(NLA=False)

    def __init__(self, config=None, **kwargs):
        # set the support range of ell1, ell2
        self.set_scale_range(config, **kwargs)
        # init line-of-sight integration config
        self.set_losint(config, **kwargs)
        # init interpolation grid
        self.set_interpolation_grid(config, **kwargs)
        # init multipole decomposition grid
        self.set_multipole_grid(config, **kwargs)
        # init intrinsic alignment model
        update_config(self.config_IA, config, **kwargs)
        
    # Binning
    def set_losint(self, config=None, **kwargs):
        """
        Set line-of-sight integration config.

        Parameters:
            config (dict)       : A configuration dict that can be used to pass in the below kwargs if
                                  desired.  This dict is allowed to have addition entries in addition
                                  to those listed below, which are ignored here. (default: None)

        Keyword arguments:
            zmin (float)        : minimum of redshift
            nzbin (int)         : number of bins for redshift
        """
        # update config
        update_config(self.config_losint, config, **kwargs)
        # source to class attributes
        self.zmin_losint  = self.config_losint['zmin']
        self.zmid_losint  = self.config_losint['zmid']
        self.nzbin_log_losint = self.config_losint['nzbin_log']
        self.nzbin_lin_losint = self.config_losint['nzbin_lin']

    def set_scale_range(self, config=None, **kwargs):
        """
        Set support range of ell1, ell2, mu.
        
        Parameters:
            config (dict)       : A configuration dict that can be used to pass in the below kwargs if
                                  desired.  This dict is allowed to have addition entries in addition
                                  to those listed below, which are ignored here. (default: None)
        
        Keyword arguments:
            ell1min (float)     : minimum of ell1 and ell2
            ell1max (float)     : maximum of ell1 and ell2
            epmu (float)        : small number to avoid the squeezed limit
        
        Description:
            Here ell1 and ell2 are the two side lengths of the triangle,
            and mu is the cosine of the **outer** angle of the triangle 
            between ell1 and ell2. Thus the other side length ell3 can be
            computed from ell1, ell2, and mu:

                ell3 = (ell1**2 + ell2**2 + 2*ell1*ell2*mu)**0.5
            
            We avoid the squeezed limit of triangle by setting the minimum
            of mu to be 1-epmu, where epmu is a small number.

            Possible config keys are:
                ell1min (float): minimum of ell1 and ell2
                ell1max (float): maximum of ell1 and ell2
                epmu (float): small number to avoid the squeezed limit
        """
        # update config
        update_config(self.config_scale, config, **kwargs)
        # fastnc args convention
        self.ell1min = self.config_scale['ell1min']
        self.ell1max = self.config_scale['ell1max']
        self.mumin    = -1.0
        self.mumax    = 1.0 - self.config_scale['epmu']

        # Multipole decomposition args convention.
        self.ellmin = 2**0.5 * self.ell1min
        self.ellmax = 2**0.5 * self.ell1max
        self.psimin = min(np.arctan2(self.ell1min, self.ell1max), np.pi/2 - np.arctan(self.ell1max/self.ell1min))
        self.psimax = np.pi/4

        # Interpolation args convention.
        # When ell, psi, mu run over the lectangular region
        # defined by the above ranges, the ranges of r, u, v
        # are also given by the ranges on ell, psi, mu.
        self.rmin = self.ellmin*min(5**-0.5, (1-self.mumax*2/5)**0.5)
        self.rmax = self.ellmax*max(2**-0.5, np.cos(self.psimin))
        self.umin = min(2**0.5*self.config_scale['epmu']**0.5, np.tan(self.psimin))
        self.umax = 1.0
        self.vmin = 0.0
        self.vmax = 1.0

    def set_interpolation_grid(self, config=None, **kwargs):
        """
        Set interpolation grid.

        Parameters:
            config (dict)       : A configuration dict that can be used to pass in the below kwargs if
                                  desired.  This dict is allowed to have addition entries in addition
                                  to those listed below, which are ignored here. (default: None)
        
        Keyword arguments:
            nrbin (int)         : number of bins for r
            nubin (int)         : number of bins for u
            nvbin (int)         : number of bins for v
            method (str)        : method for interpolation
            use_interp (bool)   : whether to use interpolation
        """
        # update config
        update_config(self.config_interp, config, **kwargs)
        # source to class attributes
        if not self.config_interp['use_interp']: return 0
        r = np.logspace(np.log10(self.rmin), np.log10(self.rmax), \
            self.config_interp['nrbin'])
        u = np.logspace(np.log10(self.umin), np.log10(self.umax), \
            self.config_interp['nubin'])
        v = np.linspace(self.vmin, self.vmax, \
            self.config_interp['nvbin'])
        # create meshgrid
        R, U, V = np.meshgrid(r, u, v, indexing='ij')
        ELL1, ELL2, ELL3 = trigutils.ruv_to_x1x2x3(R, U, V)
        # save grid
        self.r_interp = r
        self.u_interp = u
        self.v_interp = v
        self.ELL1_interp = ELL1
        self.ELL2_interp = ELL2
        self.ELL3_interp = ELL3
        # method for interpolation
        self.method_interp = self.config_interp['method']
        # place holder for interpolation function
        self.bk_interp = dict()

    def set_multipole_grid(self, config=None, **kwargs):
        """
        Set multipole decomposition grid.

        Parameters:
            config (dict)       : A configuration dict that can be used to pass in the below kwargs if
                                  desired.  This dict is allowed to have addition entries in addition
                                  to those listed below, which are ignored here. (default: None)
        
        Keyword arguments:
            nellbin (int)       : number of bins for ell
            npsibin (int)       : number of bins for psi
            nmubin (int)        : number of bins for mu
            Lmax (int)          : maximum multipole
            multipole_type (str): type of multipole decomposition
            method (str)        : method for multipole evaluation
        """
        # update config
        update_config(self.config_multipole, config, **kwargs)
        # source to class attributes
        if self.config_multipole['Lmax'] is None: return 0
        ell = np.logspace(np.log10(self.ellmin), np.log10(self.ellmax), \
            self.config_multipole['nellbin'])
        psi = loglinear(self.psimin, 1e-3, self.psimax, 50, \
            self.config_multipole['npsibin'])
        # capture the squeezed limit
        mu = 1-loglinear(1-self.mumax, 5e-2, 1-self.mumin, \
            self.config_multipole['nmubin_log'], \
            self.config_multipole['nmubin'])[::-1]
        # create meshgrid
        ELL, PSI, MU = np.meshgrid(ell, psi, mu, indexing='ij')
        ELL1, ELL2, ELL3 = trigutils.xpsimu_to_x1x2x3(ELL, PSI, MU)
        # save grid
        self.ell_multipole = ell
        self.psi_multipole = psi
        self.ELL1_multipole = ELL1
        self.ELL2_multipole = ELL2
        self.ELL3_multipole = ELL3
        # place holder for multipole decomposition
        self.bL_multipole = dict()
        # maximum multipole
        self.Lmax_multipole = self.config_multipole['Lmax']
        self.multipole_type = self.config_multipole['multipole_type']
        # For additional multipole decomposition at diagonal
        self.Lmax_multipole_diag = self.config_multipole['Lmax_diag']
        if self.Lmax_multipole_diag is None:
            self.Lmax_multipole_diag = self.Lmax_multipole
        ELL, MU = np.meshgrid(ell, mu, indexing='ij')
        self.ELL1_multipole_diag = ELL/2**0.5
        self.ELL2_multipole_diag = ELL/2**0.5
        self.ELL3_multipole_diag = ELL*(1-MU)**0.5
        self.bL_multipole_diag = dict()
        # Multipole calculator
        if self.multipole_type == 'legendre':
            self.multipole_decomposer = MultipoleLegendre(mu, \
                self.Lmax_multipole_diag, \
                method=self.config_multipole['method'])
        elif self.multipole_type == 'fourier':
            self.multipole_decomposer = MultipoleFourier(mu, \
                self.Lmax_multipole_diag, \
                method=self.config_multipole['method'])
        elif self.multipole_type == 'cosine':
            self.multipole_decomposer = MultipoleCosine(mu, \
                self.Lmax_multipole_diag, \
                method=self.config_multipole['method'])
        elif self.multipole_type == 'sine':
            self.multipole_decomposer = MultipoleSine(mu, \
                self.Lmax_multipole_diag, \
                method=self.config_multipole['method'])
        else:
            raise ValueError(f"multipole_type {multipole_type} is not supported" \
                    "supported types are 'legendre', 'fourier', 'cosine', and 'sine'")

    # Setter
    def set_cosmology(self, cosmo):
        """
        Sets cosmology. 

        Parameters:
            cosmo (astropy.cosmology): cosmology
        """
        self.cosmo = cosmo

        # compute array of chi and z
        z   = np.linspace(0, 5, 100)
        chi = self.cosmo.comoving_distance(z).value * self.cosmo.h # Mpc/h
        dzdchi = np.diff(z)/np.diff(chi)

        # spline chi <-> z
        self.z2chi = ius(z, chi)
        self.chi2z = ius(chi, z)
        self.z2dzdchi = ius(0.5*(z[1:]+z[:-1]), dzdchi, ext=1)
        self.has_changed = True

    def set_source_distribution(self, zs_list, pzs_list, sample_names=None):
        """
        Set source distribution.

        Parameters:
            zs_list (list)  : redshift array
            pzs_list (list) : probability distribution of source galaxies
        """
        # setting attributes
        self.n_sample = len(zs_list)

        # names of source samples
        if sample_names is None:
            sample_names = [str(i) for i in range(self.n_sample)]
        self.sample_names= sample_names

        # casting and shape check
        self.zs_dict = dict()
        self.pzs_dict = dict()
        for i, name in enumerate(self.sample_names):
            self.zs_dict[name] = np.asarray(zs_list[i])
            self.pzs_dict[name] = np.asarray(pzs_list[i])
            assert self.zs_dict[name].size == self.pzs_dict[name].size, \
                "zs and pzs must have the same length"

        # rise flag
        self.has_changed = True

    def set_NLA_param(self, params):
        """
        Set parameters for nonlinear alignment effect.

        Parameters:
            params (dict) : parameters for nonlinear alignment effect
        """
        self.NLA_params = params

    def set_window_function(self, window_function):
        """
        Set window function to be multiplied to the bispectrum.

        B^W(l1,l2,l3) = B(l1,l2,l3) * W(l1,l2,l3)
        """
        self.window_function = window_function

    # Redshift-bin related
    def _compute_lensing_kernel_per_sample(self, zs, pzs, nzlbin=101):
        """
        Set source distribution.

        Parameters:
            zs (array) : redshift array
            pzs (array): probability distribution of source galaxies
        """
        prefactor = 3/2 * (100/299792)**2 * self.cosmo.Om0
        if zs.size == 1:
            zl = np.linspace(self.zmin_losint, zs, nzlbin)
            chil = self.z2chi(zl)
            chis = self.z2chi(zs)
            g = prefactor*(1.-chil/chis)
        else:
            zl = np.linspace(self.zmin_losint, zs.max(), nzlbin)
            chil = self.z2chi(zl)
            chis = self.z2chi(zs)
            CHIL, CHIS = np.meshgrid(chil, chis, indexing='ij')

            # integrand
            I = np.ones_like(CHIL, dtype=float)
            I = np.divide(CHIL, CHIS, out=I, where=CHIS > CHIL)
            I = (pzs*(1-I))

            g = prefactor*np.trapz(I, zs, axis=1)/np.trapz(pzs, zs)
        return zl, chil, g

    def _compute_NLA_kernel_per_sample(self, zs, pzs, nzlbin=101):
        """
        Compute the kernel of nonlinear alignment effect.

        Parameters:
            zs (array) : redshift array
            pzs (array): probability distribution of source galaxies

        Note:
            In order for this to work, the following attributes must be set:
            - self.cosmo
            - self.z2chi
            - self.z2dzdchi
            - self.z2lgr
        """
        # model param
        AIA = self.NLA_params['AIA']
        alphaIA = self.NLA_params['alphaIA']
        z0 = self.NLA_params.get('z0',0.0)
        # constant
        c1rhocrit = 0.0134
        # compute kernel
        zsmax = zs if zs.size==1 else np.max(zs)
        zl = np.linspace(self.zmin_losint, zsmax, nzlbin)
        chil = self.z2chi(zl)
        fIA = - AIA * ((1+zl)/(1+z0))**alphaIA * c1rhocrit * self.cosmo.Om0 / self.z2lgr(zl)
        pchis = np.interp(zl, zs, pzs, left=0, right=0) * self.z2dzdchi(zl)
        norm = np.trapz(pchis, chil)
        g = fIA * pchis / norm / chil
        return zl, chil, g

    def compute_kernel(self, nzlbin=101):
        """
        Compute lensing kernel for all samples.

        Parameters:
            nzlbin (int): number of bins for lensing kernel
        """
        self.z2g_dict = dict()
        self.chi2g_dict = dict()
        for name in self.sample_names:
            z, chi, g = self._compute_lensing_kernel_per_sample(self.zs_dict[name], self.pzs_dict[name], nzlbin)
            if self.config_IA['NLA']:
                z, chi, gNLA = self._compute_NLA_kernel_per_sample(self.zs_dict[name], self.pzs_dict[name], nzlbin)
                g += gNLA
            self.z2g_dict[name] = ius(z, g, ext=1)
            self.chi2g_dict[name] = ius(chi, g, ext=1)
        self.zmax_losint = max([self.zs_dict[name].max() for name in self.sample_names])

    def get_all_sample_combinations(self):
        """
        Get all possible combinations of sample names.
        """
        combinations = []
        for i in range(self.n_sample):
            name_i = self.sample_names[i]
            for j in range(i, self.n_sample):
                name_j = self.sample_names[j]
                for k in range(j, self.n_sample):
                    name_k = self.sample_names[k]
                    combinations.append((name_i, name_j, name_k))
        return combinations

    def parse_sample_combination(self, scomb):
        """
        Parse sample combination to a list of sample names.

        Parameters:
            scomb (tuple) : a tuple of sample names (str) showing 
                            the combination of samples.

        Hint:
            If you do not have idea about the sample names, run the following code
            to get possible sample combinations.
            >>> bispectrum.get_all_sample_combinations()
            Here `bispectrum` is your bispectrum instance.
        """
        if scomb is None:
            scs = self.get_all_sample_combinations()
            assert len(scs) == 1, "specify sample_combination!"
            return scs[0]
        else:
            if np.isscalar(scomb):
                return scomb
            else:
                return tuple(scomb)
        
    # Spectra methods
    # matter power spectrum (to be implemented in subclasses)
    def matter_bispectrum(self, k1, k2, k3, z):
        """
        Compute matter bispectrum.

        Parameters:
            k1 (array) : k1 array in h/Mpc unit
            k2 (array) : k2 array in h/Mpc unit
            k3 (array) : k3 array in h/Mpc unit
            z (array)  : redshift array
        """
        raise NotImplementedError

    def get_los_kernel(self, scomb):
        # special case
        if np.isscalar(scomb):
            z = scomb
            if np.isscalar(z):
                z = np.array([z])
            chi = self.z2chi(z)
            weight = np.ones(z.size)
        else:
            # compute lensing weight, encoding geometrical dependence.
            z = loglinear(self.zmin_losint, self.zmid_losint, self.zmax_losint, \
                self.nzbin_log_losint, self.nzbin_lin_losint)
            chi = self.z2chi(z)
            weight = 1
            for name in scomb:
                weight *= self.chi2g_dict[name](chi)
            weight *= 1.0/chi*(1+z)**3
        return z, chi, weight

    # kappa bispectrum interface
    def kappa_bispectrum(self, ell1, ell2, ell3, scomb=None, \
            method='direct', **args):
        """
        Compute kappa bispectrum.

        Parameters:
            ell1 (array)  : ell1 array
            ell2 (array)  : ell2 array
            ell3 (array)  : ell3 array
            scomb (tuple) : sample combination
            method (str)  : method for computing kappa bispectrum 
                            (direct, interp, resum)
        """
        # parse sample_combination
        scomb = self.parse_sample_combination(scomb)

        if method == 'direct':
            return self.kappa_bispectrum_direct(ell1, ell2, ell3, scomb, **args)
        elif method == 'interp':
            return self.kappa_bispectrum_interp(ell1, ell2, ell3, scomb)
        elif method == 'resum':
            return self.kappa_bispectrum_resum(ell1, ell2, ell3, scomb, **args)
        else:
            raise ValueError("method must be 'direct', 'interp', or 'resum'")
        
    # direct evaluation of kappa bispectrum from matter bispectrum
    def kappa_bispectrum_direct(self, ell1, ell2, ell3, scomb=None, \
            window=True, bm=None, return_bm=False, z=None, l_shift=0.0, **args):
        """
        Compute kappa bispectrum by direct line-of-sight integration.

        Parameters:
            ell1 (array)  : ell1 array
            ell2 (array)  : ell2 array
            ell3 (array)  : ell3 array
            scomb (tuple) : sample combination
            bm (array)                : matter bispectrum, if None, it is computed
            return_bm (bool)          : return matter bispectrum if True
            args (dict)               : arguments for matter_bispectrum

        Note:
        The bm input can be used to save computation time when 
        computing kappa bispectrum for multiple sample_combinations.
        """
        # parse sample_combination
        scomb = self.parse_sample_combination(scomb)

        # check scalar
        isscalar = np.isscalar(ell1)
        if isscalar:
            ell1 = np.array([ell1])
            ell2 = np.array([ell2])
            ell3 = np.array([ell3])

        # check shape
        if ell1.shape != ell2.shape or ell1.shape != ell3.shape:
            raise ValueError("l1, l2, l3 must have the same shape")

        # save input shape
        shape = ell1.shape

        # reshape to 1d
        ell1 = ell1.ravel()
        ell2 = ell2.ravel()
        ell3 = ell3.ravel()

        # line-of-sight integration kernel
        z, chi, kernel = self.get_los_kernel(scomb)
        
        # create grids
        ELL1, Z = np.meshgrid(ell1, z, indexing='ij')
        ELL2, Z = np.meshgrid(ell2, z, indexing='ij')
        ELL3, Z = np.meshgrid(ell3, z, indexing='ij')
        CHI = self.z2chi(Z)
        # K1, K2, K3 = ELL1/CHI, ELL2/CHI, ELL3/CHI
        K1, K2, K3 = (ELL1+l_shift)/CHI, (ELL2+l_shift)/CHI, (ELL3+l_shift)/CHI

        # compute matter bispectrum
        if (bm is None) or not isinstance(scomb, tuple):
            bm = self.matter_bispectrum(K1, K2, K3, Z, **args)

        # integrand
        i = kernel * bm

        # integrate
        if i.shape[1] > 1:
            bk = np.trapz(i, chi, axis=1)
        else:
            bk = i

        # multiply window 
        if hasattr(self, 'window_function') and window:
            bk *= self.window_function(ell1, ell2, ell3)

        # reshape to the original shape
        bk = bk.reshape(shape)

        # convert to scalar if input is scalar
        if isscalar:
            bk = bk[0]

        if return_bm:
            return bk, bm
        else:
            return bk

    # interpolation
    def interpolate(self, scombs=None, **args):
        """
        Interpolate kappa bispectrum. 
        The interpolation is done in (r,u,v)-space, which is defined in M. Jarvis+2003 
        (https://arxiv.org/abs/astro-ph/0307393). See also treecorr homepage
        (https://rmjarvis.github.io/TreeCorr/_build/html/correlation3.html).

        Parameters:
            scombs (list): list of sample combinations
            args (dict): arguments for matter_bispectrum
        """
        # If sample_combinations is not given, 
        # we get all possible combinations.
        if scombs is None:
            scombs = self.get_all_sample_combinations()
        # Prepare for the interpolation
        bm = None
        grid = (np.log(self.r_interp), np.log(self.u_interp), self.v_interp)
        for sc in scombs:
            sc = self.parse_sample_combination(sc)
            bk, bm = self.kappa_bispectrum_direct(
                self.ELL1_interp, 
                self.ELL2_interp, 
                self.ELL3_interp, 
                scomb=sc,
                window=False,
                bm=bm, 
                return_bm=True, 
                **args)
            self.bk_interp[sc] = rgi(grid, np.log(bk), method=self.method_interp)

    def kappa_bispectrum_interp(self, ell1, ell2, ell3, scomb=None):
        """
        Compute kappa bispectrum by interpolation.

        Parameters:
            ell1 (array): ell1 array
            ell2 (array): ell2 array
            ell3 (array): ell3 array
        """
        scomb = self.parse_sample_combination(scomb)
        ip = self.bk_interp[scomb]
        r, u, v = trigutils.x1x2x3_to_ruv(ell1, ell2, ell3, signed=False)
        x = edge_correction(np.log(r), ip.grid[0].min(), ip.grid[0].max())
        y = edge_correction(np.log(u), ip.grid[1].min(), ip.grid[1].max())
        z = edge_correction(v, ip.grid[2].min(), ip.grid[2].max())
        bk = np.exp(ip((x,y,z)))
        # multiply window 
        if hasattr(self, 'window_function'):
            bk *= self.window_function(ell1, ell2, ell3)
        return bk

    # multipole decomposition
    def decompose(self, scombs=None, method_bispec='interp', **args):
        """
        Compute multipole decomposition of kappa bispectrum.

        Parameters:
            scombs (list)       : list of sample combinations
            method_bispec (str) : method for kappa_bispectrum
            args (dict)         : arguments for kappa_bispectrum
        """
        # If sample_combinations is not given, 
        # we get all possible combinations.
        if scombs is None:
            scombs = self.get_all_sample_combinations()
        # Compute multipole
        for sc in scombs:
            sc = self.parse_sample_combination(sc)
            b = self.kappa_bispectrum(
                    self.ELL1_multipole, 
                    self.ELL2_multipole, 
                    self.ELL3_multipole, 
                    sc, 
                    method=method_bispec, 
                    **args)
            # Compute multipoles
            L = np.arange(self.Lmax_multipole+1)
            bL = self.multipole_decomposer.decompose(b, L, axis=2)
            self.bL_multipole[sc] = bL

        # You may want to calculate higher multipole especially for 
        # diagonal elements, where ell1= ell2, corresponds to the 
        # squeezed limit bispectrum.
        if self.Lmax_multipole_diag <= self.Lmax_multipole:
            return 0
        print('Decomposing for diag.')
        for sc in scombs:
            sc = self.parse_sample_combination(sc)
            b = self.kappa_bispectrum(
                    self.ELL1_multipole_diag, 
                    self.ELL1_multipole_diag, 
                    self.ELL1_multipole_diag, 
                    sc, 
                    method=method_bispec, 
                    **args)
            # Compute multipoles
            L = np.arange(self.Lmax_multipole, self.Lmax_multipole_diag+1)
            bL = self.multipole_decomposer.decompose(b, L, axis=1)
            self.bL_multipole_diag[sc] = bL


    def kappa_bispectrum_multipole(self, L, ell, psi, scomb=None):
        """
        Compute multipole of kappa bispectrum.

        Parameters:
            L (array)     : multipole
            ell (array)   : ell array
            psi (array)   : psi array
            scomb (tuple) : sample combination
        """
        # parse sample_combination
        scomb = self.parse_sample_combination(scomb)
        # cast to array
        isscalar = np.isscalar(L)
        if isscalar:
            L = np.array([L])

        # compute multipole
        out = np.zeros((L.size,) + ell.shape)
        grid = (np.log(self.ell_multipole), np.log(self.psi_multipole))
        for i, _L in enumerate(L):
            # interpolate
            z = self.bL_multipole[scomb][_L, :, :]
            ip= rgi(grid, z, bounds_error=True)
            # convert psi to pi/2-psi if psi > pi/4
            psi = psi.copy()
            sel = np.pi/4 < psi
            psi[sel] = np.pi/2 - psi[sel]

            x = edge_correction(np.log(ell), ip.grid[0].min(), ip.grid[0].max())
            y = edge_correction(np.log(psi), ip.grid[1].min(), ip.grid[1].max())
            out[i] = ip((x, y))

        if isscalar:
            out = out[0]
            
        return out

    def kappa_bispectrum_multipole_diag(self, L, ell1, scomb=None):
        """
        Compute multipole of kappa bispectrum.

        Parameters:
            L (array)     : multipole
            ell1 (array)   : ell1 array
            scomb (tuple) : sample combination
        """
        # parse sample_combination
        scomb = self.parse_sample_combination(scomb)
        # cast to array
        isscalar = np.isscalar(L)
        if isscalar:
            L = np.array([L])

        # compute multipole
        out = np.zeros((L.size,) + ell1.shape)
        for i, _L in enumerate(L):
            # interpolate
            z = self.bL_multipole_diag[scomb][_L-self.Lmax_multipole, :]
            ip= ius(np.log(self.ell_multipole/2**0.5), z)
            out[i] = ip(np.log(ell1))

        if isscalar:
            out = out[0]
            
        return out

    def kappa_bispectrum_resum(self, ell1, ell2, ell3, scomb=None, Lmax=None):
        """
        Compute kappa bispectrum by resummation of multipoles.

        Parameters:
            L (array)     : multipole
            ell (array)   : ell array
            psi (array)   : psi array
            scomb (tuple) : sample combination
        """
        scomb = self.parse_sample_combination(scomb)
        ell, psi, mu = trigutils.x1x2x3_to_xpsimu(ell1, ell2, ell3)
        L = np.arange(Lmax or self.Lmax_multipole)
        bL = self.kappa_bispectrum_multipole(L, ell, psi, scomb=scomb)
        # pL = np.array([eval_legendre(_L, mu) for _L in L])
        _ = np.linspace(-1, 1, 100)
        pL = np.array([ius(_, eval_legendre(_L, _))(mu) for _L in L])
        out = np.sum(bL*pL, axis=0)
        return out


class BispectrumHalofit(BispectrumBase):
    """
    Bispectrum computed from halofit.
    """
    __doc__ += BispectrumBase.__doc__
    # default configs
    config_scale     = dict(ell1min=1e-1, ell1max=1e5, epmu=1e-7)
    
    def __init__(self, config=None, **kwargs):
        self.halofit = Halofit()
        super().__init__(config, **kwargs)
        self.set_baryon_param({'fb':1.0})

    def set_cosmology(self, cosmo, ns=None, sigma8=None):
        """
        Sets cosmology. 

        Parameters:
            cosmo (astropy.cosmology): cosmology
            ns (float)               : spectral index of linear power spectrum
            sigma8 (float)           : sigma8 of linear power spectrum (at z=0.0)

        Note:
            Note that the values of ns and sigma8 are set by two ways:
            1. Assigning ns and sigma8 as arguments of this method.
            2. Assigning ns and sigma8 to cosmo.meta.
        """
        super().set_cosmology(cosmo)
        # parameters for halofit
        dcosmo={'Om0': cosmo.Om0, 
                'Ode0': cosmo.Ode0,
                'ns': ns or cosmo.meta.get('n'),
                'sigma8': sigma8 or cosmo.meta.get('sigma8'), 
                'w0': cosmo.w0, 
                'wa': 0.0,
                'fnu0': 0.0} 
        self.halofit.set_cosmology(dcosmo)

    def set_pklin(self, k, pklin):
        """
        Set linear power spectrum.

        Parameters:
            k (array)    : wavenumber array
            pklin (array): linear power spectrum
        """
        self.halofit.set_pklin(k, pklin)
        self.has_changed = True

    def set_lgr(self, z, lgr):
        """
        Set linear growth rate.

        Parameters:
            z (float)  : redshift
            lgr (float): linear growth rate
        """
        self.z2lgr = ius(z, lgr, ext=1)
        self.halofit.set_lgr(z, lgr)
        self.has_changed = True

    def set_baryon_param(self, params):
        """
        Set parameter(s) of baryon

        keywords:
            fb: suppression factor relative to TNG-300
        """
        self.baryon_params = params

    def matter_bispectrum(self, k1, k2, k3, z, all_physical=True, which=['Bh1', 'Bh3']):
        b = self.halofit.get_bihalofit(k1, k2, k3, z, all_physical=all_physical, which=which)
        fb = self.baryon_params['fb']
        if fb != 0:
            Rb= self.halofit.get_Rb_bihalofit(k1, k2, k3, z)
            b*= 1.0 + fb * (Rb-1.0)
        return b

class BispectrumGilMarin(BispectrumBase):
    """
    Bispectrum computed from Gil-Marin.

    reference: https://arxiv.org/pdf/1111.4477
    """
    __doc__ += BispectrumBase.__doc__
    # default configs
    config_scale     = dict(ell1min=1e-1, ell1max=1e5, epmu=1e-7)
    
    def __init__(self, config=None, **kwargs):
        self.halofit = Halofit()
        super().__init__(config, **kwargs)

        # gil marin's parameters
        self.a1 = 0.484
        self.a2 = 3.740
        self.a3 = -0.849
        self.a4 = 0.392
        self.a5 = 1.013
        self.a6 = -0.575
        self.a7 = 0.128
        self.a8 = -0.722
        self.a9 = -0.926

    def set_cosmology(self, cosmo, ns=None, sigma8=None):
        """
        Sets cosmology. 

        Parameters:
            cosmo (astropy.cosmology): cosmology
            ns (float)               : spectral index of linear power spectrum
            sigma8 (float)           : sigma8 of linear power spectrum (at z=0.0)

        Note:
            Note that the values of ns and sigma8 are set by two ways:
            1. Assigning ns and sigma8 as arguments of this method.
            2. Assigning ns and sigma8 to cosmo.meta.
        """
        super().set_cosmology(cosmo)
        # parameters for halofit
        dcosmo={'Om0': cosmo.Om0, 
                'Ode0': cosmo.Ode0,
                'ns': ns or cosmo.meta.get('n'),
                'sigma8': sigma8 or cosmo.meta.get('sigma8'), 
                'w0': cosmo.w0, 
                'wa': 0.0,
                'fnu0': 0.0} 
        self.halofit.set_cosmology(dcosmo)

    def set_pklin(self, k, pklin):
        """
        Set linear power spectrum.

        Parameters:
            k (array)    : wavenumber array
            pklin (array): linear power spectrum
        """
        self.halofit.set_pklin(k, pklin)
        self.has_changed = True

    def set_lgr(self, z, lgr):
        """
        Set linear growth rate.

        Parameters:
            z (float)  : redshift
            lgr (float): linear growth rate
        """
        self.z2lgr = ius(z, lgr, ext=1)
        self.halofit.set_lgr(z, lgr)
        self.has_changed = True

    def set_interp(self):
        if self.has_changed:
            self.set_k2n_gilmarin()
            self.set_z2knl_gilmarin()
            self.set_z2sigma8()
            self.has_changed = False

    def set_k2n_gilmarin(self):
        """
        Get effective spectral index from Gil-Marin.
        """
        k = self.halofit.k
        pklin = self.halofit.pklin
        neff = np.diff(np.log(pklin))/np.diff(np.log(k))
        self.k2n = ius((k[1:]*k[:-1])**0.5, neff, ext=0)

    def set_z2knl_gilmarin(self):
        """
        Get non-linear scale from Gil-Marin.
        """
        # Delta without linear growht factor
        k = self.halofit.k
        pklin = self.halofit.pklin
        Delta = k**3*pklin/(2*np.pi**2)
        # growth
        z = np.linspace(0, self.halofit.z.max(), 100)
        lgr = self.z2lgr(z)
        knl = np.full_like(z, k.max())
        for i in range(z.size):
            _ = k[Delta*lgr[i]**2 < 1.0]
            if _.size > 0:
                knl[i] = _.max()
        self.z2knl = ius(z, knl, ext=0)

    def set_z2sigma8(self):
        z = np.linspace(0, self.halofit.z.max(), 100)
        s8= np.array([self.halofit.get_sigma8z(_z) for _z in z])
        self.z2sigma8 = ius(z, s8, ext=0)

    def Q3(self, n):
        return (4 - 2 ** n) / (1 + 2 ** (n + 1))

    def agm(self, k, z, knl):
        n = self.k2n(k)
        q = k / knl
        factor = (q * self.a1) ** (n + self.a2)
        sigma8 = self.z2sigma8(z)
        return (1 + sigma8**self.a6 * (0.7 * self.Q3(n) ** (1 / 2)) * factor) / (1 + factor)

    def bgm(self, k, knl):
        q = k / knl
        n = self.k2n(k)
        return (1 + 0.2 * self.a3 * (n + 3) * (q * self.a7) ** (n + 3 + self.a8)) / (1 + (q * self.a7) ** (n + 3.5 + self.a8))

    def cgm(self, k, knl):
        q = k / knl
        n = self.k2n(k)
        return (1 + 4.5 * self.a4 / (1.5 + (n + 3) ** 4) * (q * self.a5) ** (n + 3 + self.a9)) / (1 + (q * self.a5) ** (n + 3.5 + self.a9))

    def F2_eff(self, z, k1, k2, k3, knl):
        dot = (-k3 ** 2 + k1 ** 2 + k2 ** 2) / 2
        # f2 = 5 / 7 * self.agm(k1, z, knl) * self.agm(k2, z, knl) + 2 * dot ** 2 / (7 * k1 ** 2 * k2 ** 2) * self.bgm(k1, knl) * self.bgm(k2, knl) - dot * (
        #             1 / k1 ** 2 + 1 / k2 ** 2) / 2 * self.cgm(k1, knl) * self.cgm(k2, knl)
        f2 = 5 / 7 + 2 * dot ** 2 / (7 * k1 ** 2 * k2 ** 2) - dot * (1 / k1 ** 2 + 1 / k2 ** 2) / 2
        return f2

    def matter_bispectrum(self, k1, k2, k3, z, **kwargs):
        print("Note: Gilmarin ignores kwargs.")

        PNL1 = self.halofit.get_pkhalofit(k1[:,0], z[0,:]).T
        PNL2 = self.halofit.get_pkhalofit(k2[:,0], z[0,:]).T
        PNL3 = self.halofit.get_pkhalofit(k3[:,0], z[0,:]).T

        # reshape
        shape = k1.shape
        k1 = k1.ravel()
        k2 = k2.ravel()
        k3 = k3.ravel()
        z = z.ravel()
        PNL1 = PNL1.ravel()
        PNL2 = PNL2.ravel()
        PNL3 = PNL3.ravel()

        self.set_interp()
        knl = self.z2knl(z)

        bk  =  2 * (self.F2_eff(z, k1, k2, k3, knl) * PNL1 * PNL2 + \
                    self.F2_eff(z, k2, k3, k1, knl) * PNL2 * PNL3 + \
                    self.F2_eff(z, k3, k1, k2, knl) * PNL3 * PNL1)
        bk  = bk.reshape(shape)
        return bk

class BispectrumNFW1Halo(BispectrumBase):
    """
    Toy model
    """
    __doc__ += BispectrumBase.__doc__
    # default configs
    config_scale     = dict(ell1min=1e-2, ell1max=1e5, epmu=1e-7)
    
    def __init__(self, config=None, **kwargs):
        super().__init__(config, **kwargs)
        self.set_rs(rs)

    def set_rs(self, rs):
        self.rs = rs

    @classmethod
    def rhok_NFW(cls, k, rs):
        y = k*rs
        si, ci = sici(y)
        return -np.cos(y)*ci + 0.5*np.sin(y)*(np.pi-2*si)

    def kappa_bispectrum_direct(self, ell1, ell2, ell3, **args):
        rs = np.deg2rad(self.rs/60.0) # in rad
        bl = 1
        for i, _ell in enumerate([ell1, ell2, ell3]):
            bl *= self.rhok_NFW(_ell, rs)
        return bl



"""
Author: Sunao Sugiyama
Last edit  : 2024/03/25 22:02:25

This module provides the functions to manipulate the triangle configuration.
Triangle can be specified by several different ways, and this module provides
the conversion between them.

x1x2x3  (SSS) : Three side lengths parametrization. 
                Triangle orientation is ambiguous.
ruv     (RUV) : (r,u,v) parametrization introduced in M. Jarvis+2003
                Triangle orientation is expressed by sign of v.
x1x2phi (SAS) : Two side lengths and opening angle parametrization.
                Triangle orientation is expressed by sign of phi.
xpsimu        : Same as x1x2phi but ratio of x2/x1 is parametrized by tan(psi)
                and phi is reparametrized as mu=cos(phi).
"""

import numpy as np

##################################################################
def is_cyclic_permutation(x):
    N = len(x)
    
    for i in range(N):
        cyclic_permutation = np.roll(np.arange(N), i)
        if np.array_equal(x, cyclic_permutation):
            return True
    
    return False

# ruv <-> x1x2x3
def ruv_to_x1x2x3(r, u, v):
    x1 = r*(1+u*np.abs(v))
    x2 = r
    x3 = r*u
    return x1, x2, x3

def x1x2x3_to_ruv(x1, x2, x3, signed=True, all_physical=True):
    """
    x1, x2, x3 are clockwise side lengths of triangle
    """
    # Get sorted side length
    d3, d2, d1 = np.sort([x1,x2,x3], axis=0)
    r = d2
    u = d3/d2
    v = (d1-d2)/d3

    if all_physical:
        u[u>1] = 1
        v[v<0] = 0
        v[v>1] = 1

    if signed:
        # check the (d1 > d2 > d3) triangle is clockwise or not
        idx = np.argsort([x1, x2, x3], axis=0).T
        clk = [is_cyclic_permutation(_idx) for _idx in idx]
        sign = np.ones_like(clk, dtype=int)
        sign[np.logical_not(clk)] = -1
        v *= sign

    return r, u, v

# ruv <-> x1x2phi
def ruv_to_x1x2phi(r, u, v, rot=0):
    if rot == 0:
        x1, x2, x3 = (1+u*np.abs(v)), 1., u
    elif rot == 1:
        x2, x3, x1 = (1+u*np.abs(v)), 1., u
    elif rot == 2:
        x3, x1, x2 = (1+u*np.abs(v)), 1., u
    phi = np.arccos( (x1**2+x2**2-x3**2)/2/x1/x2 ) * (-np.sign(v))
    x1 *= r
    x2 *= r
    return x1, x2, phi

def x1x2phi_to_ruv(x1, x2, phi, rot=0):
    pass

# x1x2phi <-> x1x2x3
def x1x2phi_to_x1x2x3(x1, x2, phi):
    x3 = np.sqrt(x1**2 + x2**2 - 2*x1*x2*np.cos(phi))
    return x1, x2, x3

def x1x2x3_to_x1x2phi(x1, x2, x3):
    pass

# xpsimu <-> x1x2x3
def xpsimu_to_x1x2x3(x, psi, mu):
    x1 = x * np.cos(psi)
    x2 = x * np.sin(psi)
    x3 = x * np.sqrt(1 - np.sin(2*psi)*mu)
    return x1, x2, x3

def x1x2x3_to_xpsimu(x1, x2, x3):
    x = np.sqrt(x1**2 + x2**2)
    psi = np.arctan2(x2, x1)
    mu = (x1**2+x2**2-x3**2)/2/x1/x2
    return x, psi, mu


#!/usr/bin/env python
'''
Author     : Sunao Sugiyama 
Last edit  : 2024/06/09 17:37:16

Description:
coupling.py contains classes for 
the computing multipole coupling functions 
'''
import numpy as np
from scipy.special import eval_legendre
import pandas as pd
import os
import json
# fastnc modules
from .utils import sincos2angbar, npload_lock, npsavez_lock
from .integration import aint
from mpi4py import MPI

def get_cache_dir():
    # first we look for the environmental variable
    cache_dir = os.environ.get('FASTNC_CACHE_DIR')
    if cache_dir is not None:
        cache_dir = os.path.expanduser(cache_dir)
        return cache_dir
    # finally we use the default directory
    cache_dir = '~/.fastnc'
    # Expand ~ to the user's home directory
    cache_dir = os.path.expanduser(cache_dir)
    return cache_dir

def mkcachedir_if_not_exists(dirname):
    if not os.path.exists(dirname):
        print('Creating cache directory: {}'.format(dirname))
        os.makedirs(dirname, exist_ok=True)

class CacheManager:
    """
    This class manages the chache
    """
    entry_id_keys = ['name', 'Npsi', 'tol']
    def __init__(self):
        self.cache_dir = get_cache_dir()
        mkcachedir_if_not_exists(self.cache_dir)
        self.databasename = os.path.join(self.cache_dir, 'database.csv')
        self.df = self.__load_database()

    def __load_database(self):
        # Load the CSV file into memory if it exists, otherwise initialize an empty dataframe
        if os.path.exists(self.databasename):
            return pd.read_csv(self.databasename)
        else:
            return pd.DataFrame(columns=self.entry_id_keys+['filename'])

    def get_entry(self, entry_id):
        sel = [self.df[key] == entry_id[key] for key in self.entry_id_keys]
        sel = np.all(sel, axis=0)
        return self.df[sel]

    def in_database(self, entry_id):
        return not self.get_entry(entry_id).empty

    def get_entry_filename(self, entry_id):
        existing_entry = self.get_entry(entry_id)
        return os.path.join(self.cache_dir, existing_entry['filename'].values[0])

    def register_entry(self, entry_id):
        if self.in_database(entry_id):
            return 0

        # If entry does not exist, generate a new unique filename
        new_filename = self.__generate_unique_filename()
        
        # Add a new row to the dataframe with the provided Npsi, tol, and generated filename
        entry_id['filename'] = [new_filename]
        new_row = pd.DataFrame(entry_id)
        self.df = pd.concat([self.df, new_row], ignore_index=True)

        # Write the updated dataframe to the CSV file
        self.write()

    def __generate_unique_filename(self):
        # Generate a new filename that does not already exist in the dataframe
        i = 1
        while True:
            new_filename = f"coupling-{i}.npz"
            if new_filename not in self.df['filename'].values:
                return new_filename
            i += 1

    def write(self):
        # Write the current dataframe to the CSV file
        self.df.to_csv(self.databasename, index=False)

class ModeCouplingFunctionBase:
    r"""
    Compute and hold the multipole mode-coupling function between real
    and Fourier spaces.

    Parameters:
        Lmax (int)    : Maximum multipole moment.
        Mmax (int)    : Maximum angular Fourier mode.
        Npsi (int)    : Number of psi bins. Npsi is forced to be odd to capture the 
                        exact pi/2 point.
        tol (float)   : Tolerance for the numerical integration.
        verbose (bool): Whether to print verbose output.
        cache (bool)  : Whether to use cache.
    """

    # name of the mode coupling function
    name = 'ModeCouplingFunctionBase'
    
    def __init__(self, Lmax, Mmax, Npsi=200, tol=1e-5, verbose=True, cache=True):
        self.Lmax = Lmax
        self.Mmax = Mmax
        self.Npsi = int((Npsi//2)*2+1)
        self.psi = np.linspace(0, np.pi/2, Npsi)
        self.tol = tol
        self.verbose = verbose
        self.cache = cache

        # mode coupling function data
        self.data = dict()
        # load cache if exists
        self.load_cache() if self.cache else None
        # compute mode coupling function
        has_changed = self.compute()
        # save cache
        if self.cache and has_changed:
            self.save_cache()
        # correction for numerical bias
        self.correct()

    def compute(self):
        """
        This will compute the mode coupling function
        which was not found in the cache.

        Returns:
            bool: Whether the data has changed.
        """
        raise NotImplementedError

    def correct(self):
        """
        Correct the mode coupling function for numerical bias.
        """
        raise NotImplementedError

    def _get_id(self):
        return {'name':self.name, 'Npsi':self.Npsi, 'tol':self.tol}

    def save_cache(self):
        """
        Save the data to the cache.
        """
        # load database and register the data info in the database
        database = CacheManager()
        database.register_entry(self._get_id())
        # save the data to the cache
        # because the npz file only accepts string keys,
        # we convert the keys to string using json encoding
        if MPI.COMM_WORLD.Get_rank() == 0:
            filename = database.get_entry_filename(self._get_id())
            cache = {json.dumps(key): value for key, value in self.data.items()}
            npsavez_lock(filename, cache, suffix='.npz')

    def load_cache(self):
        """
        Load the data from the cache.
        """
        # load database
        database = CacheManager()
        if database.in_database(self._get_id()):
            # load the data from the cache
            filename = database.get_entry_filename(self._get_id())
            print(f'Loading cache from cache at {filename}') if self.verbose else None
            cache = npload_lock(filename, suffix='.npz')
            # decoding
            # because json encodes tuple into string which looks like a list,
            # we convert the decoded string into tuple by hand since list
            # is not hashable: it cannot be used for the key of dict.
            self.data = {tuple(json.loads(key)): value for key, value in dict(cache).items()}

    def __call__(self, L, M, psi):
        """
        Compute GLM for given L, M, and psi.

        L (array): The multipole moment.
        M (array): The angular Fourier mode.
        psi (array): The psi values.
        """
        pass

    def set_table(self, psi):
        """
        Set the table of GLM on the given psi values.
        """
        L = np.arange(self.Lmax+1)
        M = np.arange(-self.Mmax, self.Mmax+1)

        shape = psi.shape
        psi_unique, inv = np.unique(psi.ravel(), return_inverse=True)

        table = self(L, M, psi_unique)
        
        self.table_info = {}
        for i in range(L.size):
            for j in range(M.size):
                self.table_info[(L[i], M[j])] = table[i,j]
        self.table_info['inv'] = inv
        self.table_info['shape'] = shape

    def from_table(self, L, M):
        """
        Get the GLM from the table.
        """
        Lisscalar = np.isscalar(L)
        if Lisscalar:
            L = np.array([L])
        Misscalar = np.isscalar(M)
        if Misscalar:
            M = np.array([M])

        # collect todo
        todo = []
        for _L in L:
            for _M in M:
                todo.append([_L, _M])


        inv = self.table_info['inv']
        shape = self.table_info['shape']

        out = []
        for _L, _M in todo:
            o = self.table_info[(_L, _M)][inv].reshape(shape)
            out.append(o)
        out = np.array(out).reshape(L.shape+M.shape+shape)

        if Lisscalar and Misscalar:
            out = out[0,0]
        elif Lisscalar:
            out = out[0]
        elif Misscalar:
            out = out[:,0]

        return out
        
class MCF222LegendreFourier(ModeCouplingFunctionBase):
    r"""
    Mode coupling function for spin-2, spin-2, spin-2 correlation function
    and decomposition using Legendre polynomials and Fourier modes
    for Fourier-space and real-space respectively.

    .. math:
        G_LM(psi) = 4\\pi \\int_0^{\\pi} dx P_L(\\cos(x)) \\cos[2\\bar\\beta(\\psi, x) + M x]
    """
    __doc__ += ModeCouplingFunctionBase.__doc__
    
    # name of the mode coupling function
    name = 'MCF222LegendreFourier'

    def _integrand(self, x, psi, L, M):
        x, psi = np.meshgrid(x, psi, indexing='ij')
        sin2bb, cos2bb = sincos2angbar(psi, x)
        out = 4*np.pi*eval_legendre(L, np.cos(x)) * (cos2bb*np.cos(M*x) - sin2bb*np.sin(M*x))
        return out
        
    def compute(self):
        has_changed = False
        # prepare todo list
        todo = []
        for L in range(self.Lmax+1):
            for M in range(self.Mmax+1):
                todo.append([L, M])
        # compute GLM
        for L, M in todo:
            print(f'\r(L,M) = {(L,M)}/{len(todo)}', end='') if self.verbose else None
            # skip if the data already exists
            # in the cache
            if (L,M) in self.data:
                continue
            args = {'L':L, 'M':M, 'psi':self.psi}
            o, c = aint(self._integrand, 0, np.pi, 2, tol=self.tol, **args)
            self.data[(L, M)] = o
            has_changed = True
        return has_changed

    def correct(self):
        """
        Correct the mode coupling function for numerical bias.
        
        G_LM(psi) is exactly zero for L<M and psi<=pi/4. 
        However, the numerical integration may give a non-zero value.
        From my experience, the same amount of error is also present in
        G_ML if G_LM is biased. Hence we subtract the error in G_LM(psi<=pi/4)
        from G_LM(psi) and G_ML(psi).
        """
        for (L,M), data in self.data.items():
            if L>=M:
                continue
            # estimate the bias in G_LM(psi<=np.pi/4)
            bias = np.mean(data[self.psi<=np.pi/4])
            # subtract the bias
            self.data[(L,M)] -= bias
            if (M,L) in self.data:
                self.data[(M,L)] -= bias

    def __call__(self, L, M, psi):
        Lisscalar = np.isscalar(L)
        if Lisscalar:
            L = np.array([L])
        Misscalar = np.isscalar(M)
        if Misscalar:
            M = np.array([M])

        if np.any(L>self.Lmax):
            raise ValueError('L={} is larger than Lmax={}'.format(L, self.Lmax))
        if np.any(L<0):
            raise ValueError('L={} is smaller than 0'.format(L))
        if np.any(M>self.Mmax):
            raise ValueError('M={} is larger than Mmax={}'.format(M, self.Mmax))
        if np.any(M<-self.Mmax):
            raise ValueError('M={} is smaller than -Mmax={}'.format(M, self.Mmax))

        # collect todo
        todo = []
        for _L in L:
            for _M in M:
                todo.append([_L, _M])

        # Avoid the same computation
        shape = psi.shape
        psi_unique, inv = np.unique(psi.ravel(), return_inverse=True)

        out = []
        for _L, _M in todo:
            # Use the symmetry for M<0
            # G_{LM}(psi) = G_{L(-M)}(np.pi/2-psi)
            if _M>0:
                o = np.interp(psi_unique, self.psi, self.data[(_L, _M)])
            else:
                o = np.interp(np.pi/2-psi_unique, self.psi, self.data[(_L, -_M)])
            o = o[inv].reshape(shape)
            out.append(o)
        out = np.array(out).reshape(L.shape+M.shape+psi.shape)

        if Lisscalar and Misscalar:
            out = out[0,0]
        elif Lisscalar:
            out = out[0]
        elif Misscalar:
            out = out[:,0]

        return out

class MCF222FourierFourier(ModeCouplingFunctionBase):
    r"""
    Mode coupling function for spin-2, spin-2, spin-2 correlation function
    and decomposition using Fourier modes both
    for Fourier-space and real-space respectively.

    .. math:
        G_LM(\\psi) = 4\\pi \\int_0^{\\pi} dx \\cos[2\\bar\\beta(\\psi, x) + (M-L) x]

    This can be parametrized by a single integer K=M-L.
    
    ..math::
        G_K(\\psi) = 4\\pi \\int_0^{\\pi} dx \\cos[2\\bar\\beta(\\psi, x) + K x]

    """
    __doc__ += ModeCouplingFunctionBase.__doc__

    # name of the mode coupling function
    name = 'MCF222FourierFourier'
    
    def _integrand(self, x, psi, K):
        x, psi = np.meshgrid(x, psi, indexing='ij')
        sin2bb, cos2bb = sincos2angbar(psi, x)
        out = 4*np.pi*(cos2bb*np.cos(K*x) - sin2bb*np.sin(K*x))
        return out
        
    def compute(self, correct_bias=True):
        has_changed = False
        self.Kmax = self.Mmax + self.Lmax
        # compute GK
        for K in range(self.Kmax+1):
            print(f'\rK = {K}/{2*self.Kmax+1}', end='') if self.verbose else None
            # skip if the data already exists
            # in the cache
            if K in self.data:
                continue
            args = {'K':(K), 'psi':self.psi}
            o, c = aint(self._integrand, 0, np.pi, 2, tol=self.tol, **args)
            self.data[K] = o
            has_changed = True
        # Correction?
        return has_changed

    def __call__(self, L, M, psi):
        Lisscalar = np.isscalar(L)
        if Lisscalar:
            L = np.array([L])
        Misscalar = np.isscalar(M)
        if Misscalar:
            M = np.array([M])

        if np.any(L>self.Lmax):
            raise ValueError('L={} is larger than Lmax={}'.format(L, self.Lmax))
        if np.any(L<-self.Lmax):
            raise ValueError('L={} is smaller than -Lmax={}'.format(L, self.Lmax))
        if np.any(M>self.Mmax):
            raise ValueError('M={} is larger than Mmax={}'.format(M, self.Mmax))
        if np.any(M<-self.Mmax):
            raise ValueError('M={} is smaller than -Mmax={}'.format(M, self.Mmax))
        
        # collect todo
        todo = []
        for _L in L:
            for _M in M:
                todo.append([_L, _M])

        out = []
        for _L, _M in todo:
            # Use the symmetry for M<0
            # G_{K}(psi) = G_{-K}(np.pi/2-psi)
            _K = _M - _L
            if _K>0:
                o = np.interp(psi, self.psi, self.data[_K])
            else:
                o = np.interp(np.pi/2-psi, self.psi, self.data[-_K])
            out.append(o)
        out = np.array(out).reshape(L.shape+M.shape+psi.shape)

        if Lisscalar and Misscalar:
            out = out[0,0]
        elif Lisscalar:
            out = out[0]
        elif Misscalar:
            out = out[:,0]

        return out

    def correct(self):
        pass


#!/usr/bin/env python
'''
Author     : Sunao Sugiyama 
Last edit  : 2024/03/25 17:44:22

Description:
This is the integration module of fastnc.
'''
import numpy as np

def err_fnc_diff_by_norm(out1, out2):
    """
    error function for adaptive integration: 
    difference between two arrays normalized by 
    the mean of the first array. This error function 
    measures how much the integral is updated 
    from out1 to out2.
    """
    diff = np.mean(np.abs(out1-out2))
    mean = np.max(np.abs(out1))
    return diff/mean

def aint(fnc, xmin, xmax, Nx=2, axis=0, tol=1e-3, max_itern=10, err_fnc=None, verbose=False, **fnc_args):
    """
    aint (function): Adaptive integration routine.

    Parameters:
        fnc (function): Function to integrate.
        xmin (float): Lower limit of integration.
        xmax (float): Upper limit of integration.
        Nx (int): Initial number of bins.
        axis (int): Axis along which to integrate.
        tol (float): Tolerance for error.
        max_itern (int): Maximum number of iterations.
        err_fnc (function): Function to calculate the error. Default is err_fnc_diff_by_norm.
        verbose (bool): Verbose flag.

    Returns:
        out (float): Integral.
    """

    if err_fnc is None:
        err_fnc = err_fnc_diff_by_norm

    # helper function updating the integral
    def helper(x, out):
        dx = x[1]-x[0]
        xval = x + dx/2
        fval = fnc(xval, **fnc_args)
        out  = (out + np.sum(fval, axis=axis)*dx)/2.0
        x = np.sort(np.hstack([x, xval]))
        return x, out

    # initialize
    x = np.linspace(xmin, xmax, Nx+1)[:-1]
    out = np.sum(fnc(x, **fnc_args), axis=axis) * (x[1]-x[0])
    err = 1e3
    itern = 0

    # iterate until convergence
    while (err > tol) and (itern <= max_itern):
        x, out2 = helper(x, out)
        err = err_fnc(out, out2)
        out = out2
        itern += 1
        
    # warn if max iterations reached
    converged = err <= tol
    if (not converged) and verbose:
        print('Warning: max iterations reached. err={err:.2e}, tol={tol:.2e}'.format(err=err, tol=tol))

    return out, converged


#!/usr/bin/env python
'''
Author     : Sunao Sugiyama 
Last edit  : 2024/06/27 15:00:15

Description:
This is the module of fastnc, which calculate the
natural components using 2d fftlog.
'''
import numpy as np
# fastnc modules
from .twobessel import two_Bessel
from . import trigutils
from .utils import sincos2angbar, update_config, get_config_key
from .coupling import MCF222LegendreFourier, MCF222FourierFourier

class FastNaturalComponents:
    r"""
    Calculates the natural components using 2dfftlog.

    Parameters:
        config (dict)         : A configuration dict that can be used to pass in the below kwargs if
                                desired.  This dict is allowed to have addition entries in addition
                                to those listed below, which are ignored here. (default: None)
        verbose (bool)        : Whether to print the progress. (default: True)
                                
    Keyword Arguments:
        Lmax (int)            : The maximum multipole moment.
        Mmax (int)            : The maximum angular Fourier mode.
        projection (str)      : The projection of the shear. (default: 'x')
        t1 (array)            : The bin on two sides of a triangle. The theoretical prediction is given on
                                this bin if an array is given. If dlnt is also givn, this is interpreted as
                                the lower edge of the theta bins. If None, the theory is calculated on
                                FFT grid. (default: None)
        phi (array)           : The opening angle of the triangle between t1 and t2. If None is given, the 
                                class calculates the multipole of natural computes up to Mmax, which is
                                stored to the class sttributed Gamma0M, Gamma1M, Gamma2M, and Gamma3M.
                                If an array is given, this class also computes natural component on the given
                                phi bins from multipoles, which is stored to Gamma0, Gamma1, Gamma2, Gamma3.
                                (default: None)
        mu (list)             : The index of the natural component. (default: [0,1,2,3])
        dlnt (float)          : The bin width of t, on which we want to evaluate the result. (default: None)
        nu1 (float)           : The hyper parameter for 2dFFTLog. (default: 1.01)
        nu2 (float)           : The hyper parameter for 2dFFTLog. (default: 1.01)
        N_pad (int)           : The padding size for 2dFFTLog. (default: 0)
        xy (float)            : The paramater to relate real and fourier bins of FFT.
        auto (bool)           : Whether to automatically set ell1min, ell1max, and nell1bin from bispectrum. 
                                (default: True).
        ell1min (float)       : The minimum value of ell1. (default: None)
        ell1max (float)       : The maximum value of ell1. (default: None)
        nfft (int)            : The number of ell1 (or ell2) FFT grid points per one side of triangle. 
                                (default: 150)
        multipole_type (str)  : The type of multipole. (default: 'legendre')
        cache (bool)          : Whether to use cache for mode coupling function. (default: True)
        verbose (bool)        : Whether to print the progress. (default: True)

    Usage:
        >>> # Initialize the class
        >>> fnc = FastNaturalComponents(Lmax=10, Mmax=10)
        >>> # Set the bispectrum
        >>> fnc.set_bispectrum(bispectrum)
        >>> # Compute the natural components
        >>> fnc.compute()
        >>> # access to the computed natural components
        >>> print('Natural component multipole')
        >>> print(fnc.Gamma0M)
        >>> print('corresponding bins', fnc.get_bins('multipole'))
        >>> print('Natural compoenent in SAS')
        >>> print(fnc.Gamma0)
        >>> print('corresponding bins', fnc.get_bins('SAS'))
    """
    # default configuration
    projection       = 'x'
    config_multipole = {'Lmax':None, 'Mmax':None, 'Lmax_diag':None, 'multipole_type':'legendre', \
                        'use_GLM_table':False, 'cache':True}
    config_bin       = {'t1':None, 'phi':None, 'mu':[0,1,2,3], 'dlnt':None}
    config_fftlog    = {'nu1':1.01, 'nu2':1.01, 'N_pad':0, 'xy':1}
    config_fftgrid   = {'auto':True, 'ell1min':None, 'ell1max':None, 'nfft':150}
        
    def __init__(self, config=None, verbose=True, **kwargs):
        # general setup
        self.verbose = verbose
        # setup shear projection
        self.projection = get_config_key(config, 'projection', self.projection, **kwargs)
        # setup multipole
        self.set_multipole(config, **kwargs)
        # setup bin
        self.set_bin(config, **kwargs)
        # setup fft grid config
        update_config(self.config_fftgrid, config, **kwargs)
        # setup fftlog config
        update_config(self.config_fftlog, config, **kwargs)
        
    def set_multipole(self, config=None, **kwargs):
        """
        Initialize mode coupling function.

        Parameters:
            config (dict)         : A configuration dict that can be used to pass in the below kwargs if
                                    desired.  This dict is allowed to have addition entries in addition
                                    to those listed below, which are ignored here. (default: None)
                                    
        Keyword Arguments:
            Lmax (int)            : The maximum multipole moment.
            Mmax (int)            : The maximum angular Fourier mode.
            multipole_type (str)  : The type of multipole. Either 'legendre' or 'fourier'.
        """
        # update config
        update_config(self.config_multipole, config, **kwargs)
        # initialize mode coupling function
        self.Lmax = self.config_multipole['Lmax']
        self.Lmax_diag = self.config_multipole['Lmax_diag']
        if self.Lmax_diag is None:
            self.Lmax_diag = self.Lmax
        self.Mmax = self.config_multipole['Mmax']
        self.multipole_type = self.config_multipole['multipole_type']
        if self.multipole_type == 'legendre':
            self.GLM = MCF222LegendreFourier(self.Lmax_diag, self.Mmax, verbose=self.verbose, cache=self.config_multipole['cache'])
        elif self.multipole_type == 'fourier':
            self.GLM = MCF222FourierFourier(self.Lmax_diag, self.Mmax, verbose=self.verbose, cache=self.config_multipole['cache'])
        else:
            raise ValueError('Error: multipole_type={} is not expected'.format(self.multipole_type))

    def set_bispectrum(self, bispectrum):
        """
        Set the bispectrum multipoles.

        bispectrum (Bispectrum): The bispectrum object.
        """
        # check the compatibility of multipole_type
        assert self.multipole_type == bispectrum.multipole_type, \
            "multipole_type of bispectrum and FastNaturalComponents must be the same"
        # set bispectrum multipole
        self.bispectrum = bispectrum
        self.set_fftgrid()
        
    def set_bin(self, config=None, **kwargs):
        """
        Set the bin on which we want to predict theory.

        Parameters:
            config (dict)         : A configuration dict that can be used to pass in the below kwargs if
                                    desired.  This dict is allowed to have addition entries in addition
                                    to those listed below, which are ignored here. (default: None)
        
        Keyword Arguments:
            t1 (array)            : The bin on two sides of a triangle. The theoretical prediction is given on
                                    this bin if an array is given. If dlnt is also givn, this is interpreted as
                                    the lower edge of the theta bins. If None, the theory is calculated on
                                    FFT grid. (default: None)
            phi (array)           : The opening angle of the triangle between t1 and t2. If None is given, the 
                                    class calculates the multipole of natural computes up to Mmax, which is
                                    stored to the class sttributed Gamma0M, Gamma1M, Gamma2M, and Gamma3M.
                                    If an array is given, this class also computes natural component on the given
                                    phi bins from multipoles, which is stored to Gamma0, Gamma1, Gamma2, Gamma3.
                                    (default: None)
            mu (list)             : The index of the natural component. (default: [0,1,2,3])
            dlnt (float)          : The bin width of t, on which we want to evaluate the result. (default: None)
        """
        # update config
        update_config(self.config_bin, config, **kwargs)
        # Set the bin on which we want to predict theory
        self.t1 = self.config_bin['t1']
        self.phi = self.config_bin['phi']
        self.mu  = self.config_bin['mu']
        self.dnlt = self.config_bin['dlnt']
        self.tune_fftgrid = self.t1 is not None

    def set_fftgrid(self, config=None, **kwargs):
        """
        Set the grid in real and fourier space.

        Parameters:
            config (dict)         : A configuration dict that can be used to pass in the below kwargs if
                                    desired.  This dict is allowed to have addition entries in addition
                                    to those listed below, which are ignored here. (default: None)
        
        Keyword Arguments:
            auto (bool)           : Whether to automatically set ell1min, ell1max, and nell1bin from bispectrum. 
                                    (default: True).
            ell1min (float)       : The minimum value of ell1. (default: None)
            ell1max (float)       : The maximum value of ell1. (default: None)
            nfft (int)            : The number of ell1 (or ell2) FFT grid points per one side of triangle. 
                                    (default: 150)
        """
        # update config
        update_config(self.config_fftgrid, config, **kwargs)
        # Set the range of Fourier modes:
        # This is set by config or automatically
        # sourced from bispectrum support range.
        if self.config_fftgrid['auto']:
            assert hasattr(self, 'bispectrum'), \
                "bispectrum must be set to use auto mode"
            ell1min = self.bispectrum.ell1min
            ell1max = self.bispectrum.ell1max
        else:
            ell1min = self.config_fftgrid['ell1min']
            ell1max = self.config_fftgrid['ell1max']
        # 
        if self.tune_fftgrid:
            print('Tuning FFT bins...') if self.verbose else None
            # theta bin on which we want predict
            # Tune FFT grid
            
            _ = get_tuned_fftgrid(np.log(ell1min), np.log(ell1max), \
                self.config_fftgrid['nfft'], np.log(self.t1))
            self.ell1_fft = np.exp(_[0])
            self.t1_fft   = np.exp(_[1])
            self.down_sampler = _[2]
            self.config_fftlog['xy'] = np.exp(_[3])
            assert np.allclose(self.t1, self.t1_fft[self.down_sampler]), \
                "Something went wrong in tuning FFT grid\n" \
                "t1={}\nt1_fft={}".format(self.t1, self.t1_fft[self.down_sampler])
        else:
            nfft= self.config_fftgrid['nfft']
            self.ell1_fft = np.logspace(np.log10(ell1min), np.log10(ell1max), nfft)
            self.t1 = self.t1_fft = self.config_fftlog['xy']/self.ell1_fft[::-1]
            self.down_sampler = np.arange(nfft) # no downsampling
        # Allocate the same bin to second side
        self.t2 = self.t1
        self.t2_fft = self.t1_fft
        self.ell2_fft = self.ell1_fft
        # 2dFFT grid in Fourier space
        self.ELL1_FFT, self.ELL2_FFT = np.meshgrid(self.ell1_fft, self.ell2_fft, indexing='ij')
        self.ELL_FFT  = np.sqrt(self.ELL1_FFT**2 + self.ELL2_FFT**2)
        self.PSI_FFT = np.arctan2(self.ELL2_FFT, self.ELL1_FFT)
        if self.config_multipole['use_GLM_table']:
            print('Preparing GLM table...') if self.verbose else None
            self.GLM.set_table(self.PSI_FFT)
    
    def HM(self, M, bL=None, L=None, bL_diag=None, L_diag=None, **args):
        """
        Compute H_M(l1, l2 = \\sum_L (-1)^L * G_LM * b_L(l1, l2) on FFT grid.

        Parameters:
            M (int)   : The angular Fourier mode.
            bL (array): The bispectrum multipole. Defaults to None.
                        If None, it is computed using self.bispectrum.kappa_bispectrum_multipole.
                        By supplying bL, you can avoid recomputation of bL.
        """
        # Get bispectrum multipole indices, L array
        if L is None:
            L = np.arange(self.Lmax+1)
        # Get bispectrum multipole
        if bL is None:
            bL = self.bispectrum.kappa_bispectrum_multipole(
                L, self.ELL_FFT, self.PSI_FFT, **args)
        if self.config_multipole['use_GLM_table']:
            GLM = self.GLM.from_table(L, M)
        else:
            GLM = self.GLM(L, M, self.PSI_FFT)
        # Sum up GLM*bL over L
        HM = np.sum(((-1)**L*GLM.T*bL.T).T, axis=0)

        if self.Lmax_diag <= self.Lmax:
            return HM

        # Do the same for diag
        if L_diag is None:
            L_diag = np.arange(self.Lmax, self.Lmax_diag+1)
        if bL_diag is None:
            bL_diag = self.bispectrum.kappa_bispectrum_multipole_diag(
                L_diag, self.ell1_fft, **args)
        GLM_diag = self.GLM(L_diag, M, np.array([np.pi/4]))[0,:]
        HM_diag = np.sum(((-1)**L_diag*GLM_diag.T*bL_diag.T).T, axis=0)
        HM += np.diag(HM_diag)

        return HM

    def get_bins(self, bin_type='multipole', t1_unit='radian', mesh=True):
        """
        Get the bins on which the theory is computed.

        Parameters:
            bin_type (str): The type of bin. Either 'multipole' or 'SAS'.
            t1_unit (str) : The unit of t1. Either 'radian' or 'arcmin'.
            mesh (bool)   : Whether to return meshgrid or not.
        
        Returns:
            bin1 (array): The first bin.
            bin2 (array): The second bin.
            bin3 (array): The third bin.
        """
        ufactor = 1.0 if t1_unit == 'radian' else 180.0*60.0/np.pi
        if bin_type == 'multipole':
            bin1 = self.t1 * ufactor
            bin2 = self.t2 * ufactor
            bin3 = np.arange(-self.Mmax, self.Mmax+1)
        elif bin_type == 'SAS':
            bin1 = self.t1 * ufactor
            bin2 = self.t2 * ufactor
            bin3 = self.phi
        else:
            raise ValueError('Error: bin_type={} is not expected'.format(bin_type))
        if mesh:
            bin1, bin2, bin3 = np.meshgrid(bin1, bin2, bin3, indexing='ij')
        return bin1, bin2, bin3
    
    def compute(self, **args):
        """
        Compute GammaM and Gamma.

        args is a dictionary of arguments to be passed to
        self.bispectrum.kappa_bispectrum_multipole, e.g.
        sample_combination.
        """
        timer = args.pop('timer', lambda _: None)
        # 
        Lmin = args.pop('Lmin', 0)
        Lmax = args.pop('Lmax', self.Lmax)
        Mmax = args.pop('Mmax', self.Mmax)
        # natural-component multipole indices
        M = np.arange(Mmax+1)
        L = np.arange(Lmin, Lmax+1)
        # First we compute the kernel HM for all M
        bL = self.bispectrum.kappa_bispectrum_multipole(
            L, self.ELL_FFT, self.PSI_FFT, **args)
        timer('multipole')
        if self.Lmax_diag > self.Lmax:
            L_diag = np.arange(self.Lmax, self.Lmax_diag+1)
            bL_diag = self.bispectrum.kappa_bispectrum_multipole_diag(
                L_diag, self.ell1_fft, **args)
            timer('multipole diag')
        else:
            L_diag = bL_diag = None
        HM = [self.HM(_, bL=bL, L=L, L_diag=L_diag, bL_diag=bL_diag) for _ in M]
        timer('HM')

        # GammaM
        self.Gamma0M = np.zeros((self.t1.size, self.t2.size, 2*Mmax+1))
        self.Gamma1M = np.zeros((self.t1.size, self.t2.size, 2*Mmax+1))
        self.Gamma2M = np.zeros((self.t1.size, self.t2.size, 2*Mmax+1))
        self.Gamma3M = np.zeros((self.t1.size, self.t2.size, 2*Mmax+1))
        # For the sake of speed, we first loop over M
        # and then over mu. This is because the kernels of the 
        # natural-component multipoles only depends on M but not on mu.
        # This allows us to compute the kernels only once for each M.
        for _M, _HM in zip(M, HM):
            tb  = two_Bessel( \
                self.ell1_fft, self.ell2_fft, \
                _HM*self.ELL1_FFT**2*self.ELL2_FFT**2, \
                **self.config_fftlog)
            for _mu in self.mu:
                # Get (n,m) from M.
                m, n = [(_M-3,-_M-3), (-_M-1,_M-1), (_M+1,-_M-3), (_M-3,-_M+1)][_mu]
                if self.config_bin['dlnt'] is None:
                    # compute GammaM on FFT grid
                    GM = tb.two_Bessel(np.abs(m), np.abs(n))[2]
                elif self.config_bin['dlnt'] is not None:
                    # compute GammaM on FFT grid with bin-averaging effect
                    GM = tb.two_Bessel_binave(np.abs(m), np.abs(n), \
                        self.config_bin['dlnt'], self.config_bin['dlnt'])[2]
                # Apply (-1)**m and (-1)**n
                # These originate to J_m(x) = (-1)^m J_{-m}(x)
                GM *= (-1.)**m if m<0 else 1
                GM *= (-1.)**n if n<0 else 1
                # normalization
                GM /= (2*np.pi)**3
                # downsample
                GM = GM[np.ix_(self.down_sampler, self.down_sampler)]

                # Store
                if _mu == 0:
                    self.Gamma0M[:,:, _M+Mmax] = GM
                    self.Gamma0M[:,:,-_M+Mmax] = GM.T
                elif _mu == 1:
                    self.Gamma1M[:,:, _M+Mmax] = GM.T
                    self.Gamma1M[:,:,-_M+Mmax] = GM
                elif _mu == 2:
                    self.Gamma2M[:,:, _M+Mmax] = GM
                    self.Gamma3M[:,:,-_M+Mmax] = GM.T
                elif _mu == 3:
                    self.Gamma3M[:,:, _M+Mmax] = GM
                    self.Gamma2M[:,:,-_M+Mmax] = GM.T
        timer('GammaM')

        if self.phi is None:
            return
        M = np.arange(-Mmax, Mmax+1)
        expMphi = np.exp(1j*M[:,None]*self.phi[None,:])
        # resum multipoles
        self.Gamma0 = np.dot(self.Gamma0M, expMphi)/(2*np.pi)
        self.Gamma1 = np.dot(self.Gamma1M, expMphi)/(2*np.pi)
        self.Gamma2 = np.dot(self.Gamma2M, expMphi)/(2*np.pi)
        self.Gamma3 = np.dot(self.Gamma3M, expMphi)/(2*np.pi)
        timer('Gamma')
        # change shear projection
        self._change_shear_projection('x', self.projection)
        timer('projection')
    
    def _change_shear_projection(self, dept, dest):
        """
        Change the shear projection.

        Parameters:
            dept (str): The current shear projection.
            dest (str): The destination shear projection.
        """
        if self.verbose and dept != dest:
            print('changing shear projection from {} to {}'.format(dept, dest))
        # attributes are 1d arrays, so we cast them to 3d arrays
        T1  = self.t1[:,None,None]
        T2  = self.t2[None,:,None]
        PHI = self.phi[None,None,:]
        # Convert
        if dept == dest:
            return
        elif dept == 'x' and dest == 'cent':
            self.Gamma0 *= x2cent(0, T1, T2, PHI)
            self.Gamma1 *= x2cent(1, T1, T2, PHI)
            self.Gamma2 *= x2cent(2, T1, T2, PHI)
            self.Gamma3 *= x2cent(3, T1, T2, PHI)
        elif dept == 'cent' and dest == 'x':
            self.Gamma0 /= ortho2cent(0, T1, T2, PHI)
            self.Gamma1 /= ortho2cent(1, T1, T2, PHI)
            self.Gamma2 /= ortho2cent(2, T1, T2, PHI)
            self.Gamma3 /= ortho2cent(3, T1, T2, PHI)
        elif dept == 'x' and dest == 'ortho':
            self.Gamma0 *= x2ortho(0, T1, T2, PHI)
            self.Gamma1 *= x2ortho(1, T1, T2, PHI)
            self.Gamma2 *= x2ortho(2, T1, T2, PHI)
            self.Gamma3 *= x2ortho(3, T1, T2, PHI)
        elif dept == 'ortho' and dest == 'x':
            self.Gamma0 /= x2ortho(0, T1, T2, PHI)
            self.Gamma1 /= x2ortho(1, T1, T2, PHI)
            self.Gamma2 /= x2ortho(2, T1, T2, PHI)
            self.Gamma3 /= x2ortho(3, T1, T2, PHI)
        elif dept == 'ortho' and dest == 'cent':
            self.Gamma0 *= x2cent(0, T1, T2, PHI)
            self.Gamma1 *= x2cent(1, T1, T2, PHI)
            self.Gamma2 *= x2cent(2, T1, T2, PHI)
            self.Gamma3 *= x2cent(3, T1, T2, PHI)
        elif dept == 'cent' and dest == 'ortho':
            self.Gamma0 /= x2cent(0, T1, T2, PHI)
            self.Gamma1 /= x2cent(1, T1, T2, PHI)
            self.Gamma2 /= x2cent(2, T1, T2, PHI)
            self.Gamma3 /= x2cent(3, T1, T2, PHI)
        else:
            raise ValueError('Error: dept={} and dest={} is not expected'.format(dept, dest))


# phase factors to convert between different projections
def x2ortho(mu, t1, t2, phi):
    """
    Convert x-projection to orthocenter-projection.

    Parameters:
        mu (int)    : The index of the natural component.
        t1 (array)  : The value of t1.
        t2 (array)  : The value of t2.
        phi (array) : The value of phi.
    """
    # Compute prefactor
    sin2pb, cos2pb = sincos2angbar(np.arctan2(t2, t1), np.pi-phi)
    if mu==0 or mu==1 or mu==2:
        out = cos2pb - 1j*sin2pb
    elif mu==3:
        out = cos2pb + 1j*sin2pb
    return out

def ortho2cent(mu, t1, t2, phi):
    """
    Convert orthocenter-projection to centroid-projection.

    Parameters:
        mu (int)    : The index of the natural component.
        t1 (array)  : The value of t1.
        t2 (array)  : The value of t2.
        phi (array) : The value of phi.
    """
    NotImplementedError('This function is not validated yet')
    t1, t2, t3 = trigutils.x1x2phi_to_x1x2x3(t1, t2, phi)

    def temp(t1, t2, t3):
        phi3 = np.arccos( (t1**2+t2**2-t3**2)/2/t1/t2 )
        cos2psi = ((t2**2-t1**2)**2 - 4*t1**2*t2**2*np.sin(phi3)**2)/4.0
        sin2psi = (t2**2-t1**2) * t1*t2 * np.sin(phi3)
        norm = np.sqrt(cos2psi**2 + sin2psi**2)
        exp2psi = cos2psi/norm + 1j*sin2psi/norm
        return exp2psi

    exp2psi3 = temp(t1, t2, t3)
    exp2psi1 = temp(t2, t3, t1)
    exp2psi2 = temp(t3, t1, t2)

    out = 1
    for j, phase in enumerate([1.0, exp2psi1, exp2psi2, exp2psi3]):
        if j == mu:
            out *= phase
        else:
            out *= np.conj(phase)

    return out

def x2cent(mu, t1, t2, phi):
    """
    Convert x-projection to centroid-projection.

    Parameters:
        mu (int)    : The index of the natural component.
        t1 (array)  : The value of t1.
        t2 (array)  : The value of t2.
        phi (array) : The value of phi.

    References:
        Equations between Eq. (15) and (16) 
        of https://arxiv.org/abs/2309.08601
    """
    v = t1+t2*np.exp(-1j*phi)
    q1 = v/np.conj(v)
    v = -2*t1+t2*np.exp(-1j*phi)
    q2 = v/np.conj(v)
    v = t1-2*t2*np.exp(-1j*phi)
    q3 = v/np.conj(v)

    if mu==0:
        out = q1*q2*q3 * np.exp(3j*phi)
    elif mu==1:
        out = np.conj(q1)*q2*q3 * np.exp(1j*phi)
    elif mu==2:
        out = q1*np.conj(q2)*q3 * np.exp(3j*phi)
    elif mu==3:
        out = q1*q2*np.conj(q3) * np.exp(-1j*phi)
    else:
        raise ValueError('Error: mu={} is not expected'.format(mu))

    return out

def tune_fft_grid_size(lmin, lmax, nfft_min, dt, log=False):
    """
    This function tunes the size of FFT grid so that
    the bin size to be an integer multiple of dt, 
    and also number of bins to be greater than nfft_min.

    Parameters:
        lmin    (float): The minimum value of t.
        lmax    (float): The maximum value of t.
        nfft_min(int)  : The minimum number of FFT grid.
        dt      (float): The bin width of t, on which 
                        we want to evaluate the result

    Derivation:
        We parametrize the FFT bin size as
            dl = dt/nskip
        where nskip is an integer to be tuned
        Using this, the number of FFT bins is given by
            nfft = (lmax-lmin)/dl + 1
        The requirement on nfft is
            nfft >= nfft_min
        Also, nfft must be an integer, so by defining
            (lmax-lmin)/dl+1 = nfft + ep
        where nfft is the integer part of the lhs and 
        ep is the fractional part, we require
            0 <= ep < 1
        After tuning nskip, we also need to tune the range as
            lmax = lmax - dl*ep/2
            lmin = lmin + dl*ep/2

        The nskip can be found by
            nskip = ceil( (nfft_min-1)*dt/(lmax-lmin) )
        Once nskip is chosen, the other parameters can be derived
        as follows
            dl = dt/nskip
            nfft = floor( (lmax-lmin)/dl+1 )
            ep = (lmax-lmin)/dl - nfft
            lmax = lmax - dl*ep/2
            lmin = lmin + dl*ep/2

        With this output you can make FFT grid by
            np.linspace(lmin, lmax, nfft)
    """
    nskip = np.ceil( (nfft_min-1)*dt/(lmax-lmin) ).astype(int)
    dl = dt/nskip
    nfft = np.floor( (lmax-lmin)/dl+1 ).astype(int)
    ep = (lmax-lmin)/dl+1 - nfft
    lmax = lmax - dl*ep/2
    lmin = lmin + dl*ep/2
    if nfft%2 != 0:
        nfft -= 1
        lmax -= dl/2
        lmin += dl/2
    return lmin, lmax, nfft, nskip

def tune_fft_real_bin(l, t_pivot):
    """
    This tunes the position of fft grid in real
    space so that we can easily get the desired 
    bin `t` just by downsampling the t_fft.

    Parameters:
        l       (array): The fft grid in fourier space.
        t_pivot (float): The pivot value of t.
    
    Derivation:
        The fft grid in fourier and real space is 
        expressed by
            l_i = l_0 + i dl
            t_i = t_0 + i dt
        where dl = dt and i=0,1,...,N-1. 
        We can relate them as
            t_0(p)  = p - l_{N-1}
                    = p - l_max
        This is equivalent with
            t_i(p)  = p - l_{N-1} + idt
                    = p - l_{N-1-i}
        
        Consider we want one of the real-space
        fft grid to fall on t_pivot by nicely choosing p.
        First, we consider p=0, and where in fft grid
        t_pivot falls on. Because p is not tuned, there
        is no garantee that t_pivot falls on the grid.
        We denote this residual in grid index by ep.
            t_pivot = t_(i+ep)(0)
                    = -l_max + (i+ep)dt
        Then this ep is given as the fractional part of
            ep = frac[ (t_pivot+l_max)/dt ]
        Using this ep value, we can then chose p value
            t_pivot = ep*dt - l_max + i*dt
            p = ep*dt
    """
    dl = l[1]-l[0]
    ep = np.modf( (t_pivot+l[-1])/dl )[0]
    p = ep*dl
    return p

def get_tuned_fftgrid(lmin, lmax, nfft_min, t):
    """
    This function returns the tuned FFT grid in
    fourier space and the corresponding real-space
    grid index.

    Parameters:
        lmin    (float): The minimum value of t.
        lmax    (float): The maximum value of t.
        nfft_min(int)  : The minimum number of FFT grid.
        t       (float): The pivot value of t.
    """
    dt = np.diff(t)[0]
    lmin, lmax, nfft, nskip = tune_fft_grid_size(lmin, lmax, nfft_min, dt)
    l_fft = np.linspace(lmin, lmax, nfft)
    p = tune_fft_real_bin(l_fft, t[0])
    t_fft = p-l_fft[::-1]
    start = np.argmin(np.abs(t_fft-t[0]))
    sel = start + nskip*np.arange(t.size)
    return l_fft, t_fft, sel, p
